import math

import coopgame.characterDevelopment.character as char
from coopgame.characterDevelopment import stats as stat
import logging

logger = logging.getLogger(__name__)

HP = 'HP'
MAX_HP = 'MAX_HP'

def current_hp(character: char.Character,
               hp_stat_name: str = HP) -> float:
    return character.stats.State.resolve_stat(stat_name=hp_stat_name).val

def max_hp(character: char.Character,
           max_hp_stat_name: str = MAX_HP) ->float:
    return character.stats.State.resolve_stat(stat_name=max_hp_stat_name).val

def alive(character: char.Character,
          hp_stat_name: str = HP) -> bool:
    return not dead(character=character,
                    hp_stat_name=hp_stat_name)

def dead(character: char.Character,
          hp_stat_name: str = HP) -> bool:
    current = current_hp(character=character, hp_stat_name=hp_stat_name)
    return math.isclose(current, 0)


def heal(character: char.Character,
         amount: float = None,
         hp_stat_name: str = HP,
         max_hp_stat_name: str = MAX_HP
         ):
    if amount is None:
        amount = missing_hp(character=character,
                            hp_stat_name=hp_stat_name,
                            max_hp_stat_name=max_hp_stat_name)

    character.stats.register_events(events=[stat.DeltaStatStatsPackageStateEvent(stat_name=hp_stat_name, delta=amount)])

    log_txt = f"Character: {character.name} gained {amount} of hp!"
    current = current_hp(character=character, hp_stat_name=hp_stat_name)
    max = max_hp(character=character, max_hp_stat_name=max_hp_stat_name)
    if math.isclose(max, current):
        log_txt += " Character is at full health!"
    logger.info(log_txt)
    return character

def damage(character: char.Character,
           amount: float,
           hp_stat_name: str = HP):
    character.stats.register_events(events=[stat.DeltaStatStatsPackageStateEvent(stat_name=hp_stat_name, delta=-amount)])

    current = current_hp(character=character, hp_stat_name=hp_stat_name)

    log_txt = f"Character: {character.name} lost {amount} of hp! ({current} remaining)"
    if math.isclose(current, 0):
        log_txt += " Character was killed..."
    logger.info(log_txt)

    return character

def missing_hp(character: char.Character,
               hp_stat_name: str = HP,
               max_hp_stat_name: str = MAX_HP):
    current = current_hp(character, hp_stat_name=hp_stat_name)
    max = max_hp(character, max_hp_stat_name=max_hp_stat_name)

    return max - current

def kill(character: char.Character,
           hp_stat_name: str = HP
         ):
    current = current_hp(character=character,
                         hp_stat_name=hp_stat_name)
    damage(character=character,
           hp_stat_name=hp_stat_name,
           amount=current)
    return character

if __name__ == "__main__":
    import pprint

    default_stats = [
        stat.StatState(HP, val=1000, meta=stat.StatMeta(0, 1000)),
        stat.StatState(MAX_HP, val=1000, meta=stat.StatMeta()),
    ]

    me = char.Character(args=char.CharacterGeneratorArgs(
                            name_provider="Ralphius",
                            init_stats_package=stat.StatsPackageState(
                                base_stats=stat.StatStates({x.stat_name: x for x in default_stats})
                            )
                        )
    )

    pprint.pprint(me.stats.State.Summary)
    damage(me, 19)
    pprint.pprint(me.stats.State.Summary)
    heal(me, 10)
    pprint.pprint(me.stats.State.Summary)
    heal(me)
    pprint.pprint(me.stats.State.Summary)

    kill(me)
    pprint.pprint(me.stats.State.Summary)





from coopgame.gameTemplate import GameTemplate
from cooptools.colors import Color
import pygame
from coopstructs.vectors import Vector2
from noise import snoise2, pnoise3
import math
from coopbugger.monitoredclass import MonitoredClass
import logging
from coopstructs.toggles import BooleanToggleable
import numpy as np
import cooptools.marchingSquares as msqrs
from coopgame.tilemap import TileSet, Tilemap


REZ_X = 50
REZ_Y = 50
MAX_SIZE = 20


class Particle():
    def __init__(self, pos: Vector2, value, size, color = None):
        self.pos = pos
        self.size = size
        self.thickness = 1
        self.value = value
        self.color = color

    def display(self, screen: pygame.Surface, color = None):
        if color is None:
            color = self.color
        pygame.draw.circle(screen, color, (self.pos.x, self.pos.y), self.size, self.thickness)

class MarchingSquaresGame(GameTemplate, MonitoredClass):
    def __init__(self):
        GameTemplate.__init__(self, fullscreen=False)
        MonitoredClass.__init__(self)
        self.cols = REZ_X
        self.rows = REZ_Y

        self.xoff = 0
        self.yoff = 0
        self.zoff = 0
        self.xyincrement = .002
        self.zincrement = 0

        self.noise_scale = max(self.rows, self.cols)/5
        self.particles = []

        self._blitables = {}
        self.show_overlap = False
        self.show_lines = BooleanToggleable(default=False)
        self._init_blitables()


        self.create_particles()

        self.line_surface_layers = []
        self.lines_surface = None
        self.lines_surface_offset = Vector2(0, 0)

        self._outline_tileset = None
        self.init_tileset()
        self._outline_tilemap = None

    def _init_blitables(self):
        for ii in range(1, MAX_SIZE):
            blitable = pygame.Surface(size=(ii * 2, ii*2))
            if self.show_overlap:
                blitable.set_colorkey(Color.BLACK.value)
            color = (0, 0, int(255 * (ii / MAX_SIZE)))
            pygame.draw.circle(blitable, color, (ii, ii), ii, 1)
            self._blitables[ii] = blitable

    @MonitoredClass.timer
    def draw(self, frames:int, debug_mode: bool = False):
        self.xoff += self.xyincrement
        self.yoff += self.xyincrement
        self.zoff += self.zincrement

        self.update_particles()

        for row in self.particles:
            for particle, color in row:
                self.screen.blit(self._blitables[particle.size], (particle.focal_point.x, particle.focal_point.y))

        if self.show_lines.value:
            self.draw_lines()

        if frames % 100 == 0:
            print(self.tracked_time)

    @MonitoredClass.timer
    def build_line_surfaces(self):
        if self.zincrement != 0:
            return None

        surfaces = []

        # Move existing surfaces
        for surface, offset in self.line_surface_layers:
            offset += Vector2(-1, -1) / self.noise_scale * ((REZ_Y + REZ_X) / 2)
            if offset.x > -self.screen.get_width() and offset.y > -self.screen.get_height():
                surfaces.append((surface, offset))

        # create new surface
        lines_surface = pygame.Surface((self.screen.get_width(), self.screen.get_height()), pygame.SRCALPHA)
        lines_surface = lines_surface.convert_alpha()
        if len(self.line_surface_layers) == 0:
            self.draw_lines_to_surface(lines_surface)
            surfaces.append((lines_surface, Vector2(0, 0)))
        elif self.line_surface_layers[-1][1].length() >= (Vector2(REZ_X, REZ_Y).length()):
            self.draw_lines_to_surface(lines_surface, only_last=True)
            surfaces.append((lines_surface, Vector2(0, 0)))

        return surfaces


    def handle_left_mouse_down(self, pressed_keys, mouse_pos: Vector2):
        self.show_overlap = not self.show_overlap
        self._init_blitables()

    def handle_right_mouse_up(self, pressed_keys, mouse_pos_as_vector: Vector2):
        self.show_lines.toggle()

    @MonitoredClass.timer
    def update_particles(self):

        if self.particles is None or len(self.particles) == 0:
            self.create_particles()

        octaves = 6
        persistence = .5
        lacunarity = 2


        for ii in range(0, self.cols):
            # self.particles.append([])
            for jj in range(0, self.rows):
                value = pnoise3(
                    (ii) / self.noise_scale + self.xoff,
                    (jj) / self.noise_scale + self.yoff,
                    self.zoff,
                    octaves=octaves,
                    persistence=persistence,
                    lacunarity=lacunarity,
                    repeatx=1024,
                    repeaty=1024,
                    base=0)
                color = (0, 0, int(255 * (value + 1 / 2)))

                self.particles[ii][jj][0].value = value
                self.particles[ii][jj][0].size = max(int((value + 1 / 2) * MAX_SIZE), 1)
                self.particles[ii][jj][0].color = color



    @MonitoredClass.timer
    def create_particles(self):
        self.particles.clear()

        for ii in range(0, self.cols):
            self.particles.append([])
            for jj in range(0, self.rows):
                self.particles[ii].append(
                    (Particle(Vector2(ii * self.sq_width, jj * self.sq_height), 0, 1), Color.WHITE))

    @MonitoredClass.timer
    def get_marching_squares_state_array(self):
        # get an array of the value of particles
        part_value_arr = np.array(
            [[particle[0].value for particle in row] for row in self.particles]
        ).transpose()

        # get the state of each square
        line_defs = msqrs.marching_squares(part_value_arr, qualifier=lambda x: math.ceil(x) > 0, ret_type=msqrs.MarchingSquaresReturnType.STATE)

        # convert the dict to an np array containing states that map to tileset
        max_0 = max(x[0] for x in line_defs.keys())
        max_1 = max(x[1] for x in line_defs.keys())
        new_arr = np.zeros((max_0 + 1, max_1 + 1), dtype=int)
        for key, val in line_defs.items():
            new_arr[key[0]][key[1]] = val

        return new_arr

    @MonitoredClass.timer
    def draw_lines_to_surface(self, surface: pygame.Surface, only_last=False):
        counter = 0

        new_arr = self.get_marching_squares_state_array()

        # create a tilemap
        self._outline_tilemap = Tilemap(self._outline_tileset, map=new_arr, surface_size=(self.screen.get_width(), self.screen.get_height()))

        # blit tilemap to the screen
        self.screen.blit(self._outline_tilemap.image, self._outline_tilemap.rect)

        logging.debug(f"{counter} lines drawn")

    @MonitoredClass.timer
    def draw_lines(self):
        if self.line_surface_layers is not None and len(self.line_surface_layers) > 0:
            print(len(self.line_surface_layers))
            for surface, offset in self.line_surface_layers:
                self.screen.blit(surface, (offset.x, offset.y))
        else:
            self.draw_lines_to_surface(self.screen)

    def init_tileset(self):

        tiles = []

        mdpts = msqrs.square_midpoints(0, 0, self.sq_width, self.sq_height)

        for state, condition in msqrs.state_mapper.items():
            new_surf = pygame.Surface((self.sq_width, self.sq_height))
            new_surf.fill(Color.HOT_PINK.value)
            new_surf.set_colorkey(Color.HOT_PINK.value)
            new_surf.convert_alpha()
            for line in condition:
                pygame.draw.line(new_surf, Color.WHITE.value, mdpts[line[0]], mdpts[line[1]])
            tiles.append(new_surf)

        self._outline_tileset = TileSet(tiles=tiles)

    @property
    def sq_width(self):
        return self.screen_width / REZ_X

    @property
    def sq_height(self):
        return self.screen_height / REZ_Y

if __name__ == "__main__":
    game = MarchingSquaresGame()
    game.main()


from typing import List
from coopstructs.geometry.vectors.vectorN import VectorN, Vector2, Vector3
import numpy as np
import coopgame.pygamehelpers as help
from coopstructs.geometry import Rectangle
import math
from coopgame.renderedObjectHandling.objectOrientationArgs import ObjectOrientationArgs
from coopgame.renderedObjectHandling.objectOrientationMatrixFactory import object_orientation_matrix
from coopgame.renderedObjectHandling.renderedViewArgs import RenderedViewArgs

class DrawMatrixState:

    def __init__(self,
                 cameras_vision_matrix=None,
                 perspective_matrix=None):
        self.cameras_vision_matrix = cameras_vision_matrix
        self.perspective_matrix = perspective_matrix

    def set_matricies(self, rendered_view_args: RenderedViewArgs):
        self.cameras_vision_matrix = self._create_cameras_vision_matrix(camera_orientation=rendered_view_args.camera_orientation)
        self.perspective_matrix = self._create_viewport_projection_matrix(near_plane_dist=rendered_view_args.near_plane_dist,
                                                                        far_plane_dist=rendered_view_args.far_plane_dist,
                                                                        field_of_view_rads=rendered_view_args.field_of_view_rads)
    @staticmethod
    def _create_cameras_vision_matrix(camera_orientation: ObjectOrientationArgs):
        return object_orientation_matrix(camera_orientation)

    @staticmethod
    def _create_viewport_projection_matrix(near_plane_dist: float = None,
                                           far_plane_dist: float = None,
                                           field_of_view_rads: float = None):
        # https://www.scratchapixel.com/lessons/3d-basic-rendering/perspective-and-orthographic-projection-matrix/building-basic-perspective-projection-matrix

        if near_plane_dist is None:
            near_plane_dist = 0

        if far_plane_dist is None:
            far_plane_dist = 1000000

        if field_of_view_rads is None:
            field_of_view_rads = math.pi / 2

        # field_of_view_degrees = field_of_view_rads * 180 / math.pi
        S = 1 / math.tan(field_of_view_rads / 2)

        a = - far_plane_dist / (far_plane_dist - near_plane_dist)
        b = - (far_plane_dist * near_plane_dist) / (far_plane_dist - near_plane_dist)

        # Transpose because we assume we apply S * x (matrix on left)
        return np.array([[S, 0, 0, 0],
                         [0, S, 0, 0],
                         [0, 0, a, b],
                         [0, 0, -1, 0]])


class MultiPerspectivePointConverter:

    def __init__(self):

        self._real_world_points = None
        self._camera_view_world_points = None
        self._projection_points = None
        self._viewport_points = None

    @property
    def real_world_points(self):
        return self._real_world_points

    @property
    def camera_view_world_points(self):
        return self._camera_view_world_points

    @property
    def viewport_points(self):
        return self._viewport_points

    @property
    def projection_points(self):
        return self._projection_points

    def convert_real_to_viewport(self,
                real_points: List[Vector3],
                matrix_state: DrawMatrixState,
                game_area_rect: Rectangle,
                ):
        self._real_world_points = real_points
        self._camera_view_world_points = CoordinateConverter.world_to_scaled_points(self.real_world_points, matrix_state.cameras_vision_matrix)

        self._projection_points = CoordinateConverter.scale_points(self._camera_view_world_points, matrix_state.perspective_matrix)

        self._viewport_points = [Vector2(min(game_area_rect.width - 1, ((point.x + 1) * 0.5 * game_area_rect.width)),
                                         min(game_area_rect.height - 1, ((1 - (point.y + 1) * 0.5) * game_area_rect.height)
                                             )) for point in self._projection_points]
        return self._viewport_points

    def convert_viewport_to_real(self,
                                  viewport_points: List[Vector2],
                                  game_area_rect: Rectangle,
                                  world_plane_points: List[Vector3],
                                  matrix_state: DrawMatrixState,
                                  ):
        self._viewport_points = viewport_points
        self._camera_view_world_points = CoordinateConverter.viewport_points_on_a_world_plane(viewport_points=viewport_points,
                                                                                           game_area_rect=game_area_rect,
                                                                                           world_plane_points=world_plane_points)


        inv_camera_mat = np.linalg.inv(matrix_state.cameras_vision_matrix)
        self._real_world_points = CoordinateConverter.scale_points(self._camera_view_world_points, inv_camera_mat)


class CoordinateConverter:
    @classmethod
    def world_to_scaled_points(cls, world_points: List[VectorN], matrix=None):
        return cls.scale_points(world_points, matrix=matrix)

    @classmethod
    def mouse_viewport_pos_as_vector(cls, viewport: Rectangle):
        return help.game_area_coords_from_parent_coords(parent_coords=help.mouse_pos_as_vector(),
                                                        game_area_surface_rectangle=viewport)

    @classmethod
    def _scaled_rectangle_points(cls, rectangle: Rectangle, matrix=None):
        rec_points = [Vector2(point[0], point[1]) for point in rectangle.points_tuple()]
        return cls.scale_points(rec_points,
                                  matrix=matrix)

    @classmethod
    def mouse_world_pos(cls, game_area_rect: Rectangle, draw_scale_matrix=None):
        points_on_plane = cls._scaled_rectangle_points(game_area_rect, draw_scale_matrix)
        mouse_viewport_pos = cls.mouse_viewport_pos_as_vector(game_area_rect)

        mouse_pos_in_world = cls.viewport_points_on_a_world_plane([mouse_viewport_pos],
                                                                  game_area_rect=game_area_rect,
                                                                  world_plane_points=points_on_plane)[0]
        return mouse_pos_in_world

    @classmethod
    def viewport_points_on_a_world_plane(cls,
                                         viewport_points: List[Vector2],
                                         game_area_rect: Rectangle,
                                         world_plane_points: List[Vector3] = None) -> List[Vector3]:

        if world_plane_points is None:
            viewport_points = game_area_rect.points_tuple()
            world_plane_points = [Vector3(point[0], point[1], 0) for point in viewport_points]
        elif len(world_plane_points) < 3:
            raise ValueError(f"plane needs to be defined by at least 3 points but {len(world_plane_points)} were given")

        plane_point = cls.points_projected_to_plane(viewport_points, world_plane_points[0], world_plane_points[1], world_plane_points[2])

        return plane_point

    @staticmethod
    def points_projected_to_plane(viewport_points: List[Vector2],
                                      plane_point_1: Vector3,
                                      plane_point_2: Vector3,
                                      plane_point_3: Vector3) -> List[Vector3]:

        vec1 = plane_point_1 - plane_point_2
        vec2 = plane_point_3 - plane_point_2

        normal = np.cross(vec1.as_tuple(), vec2.as_tuple())
        a = normal[0]
        b = normal[1]
        c = normal[2]
        d = a * plane_point_1.x + b * plane_point_1.y + c * plane_point_1.z

        z_val_lam = lambda point: (d - a * point.x - b * point.y) / c
        # z_val = (d - a * viewport_point.x - b * viewport_point.y) / c

        return [Vector3(point.x, point.y, z_val_lam(point)) for point in viewport_points]



    @staticmethod
    def apply_matrix_to_array(array: np.array, matrix=None, round_digits: int = 10):
        if matrix is None:
            matrix = np.identity(4)

        # Multiply the points by the transform matrix for drawing'''
        transformed_points = matrix.dot(
            np.transpose(array))  # Transpose the points to appropriately multiply

        ret = np.transpose(transformed_points)

        if ret.size == 4:
            ret = ret / ret[3]
        else:
            ret = ret / ret[:, 3].reshape((ret.shape[0], 1))
        return ret.round(round_digits)

    @classmethod
    def scale_points(cls, points: List[VectorN], matrix=None) -> List[VectorN]:

        if len(points) == 0 or matrix is None:
            return points

        '''Convert the point to a 4-dim point for multiplication'''
        normal_array = [(point.x, point.y, point.z if point.z else 0, 1) for point in points]

        # Scale the array
        scaled_array = cls.apply_matrix_to_array(normal_array, matrix=matrix)

        # Convert back to points
        scaled_points = [VectorN({'x': point[0], 'y':point[1], 'z': point[2]}) for point in scaled_array]
        for ii, point in enumerate(points):
            if type(point) == Vector2:
                scaled_points[ii] = Vector2(scaled_points[ii].x, scaled_points[ii].y)
            elif type(point) == Vector3:
                scaled_points[ii] = Vector3(scaled_points[ii].x, scaled_points[ii].y, scaled_points[ii].z)


        return scaled_points

if __name__ == "__main__":
    from pprint import pprint
    real_points = [
        Vector3(1, 1, 3),
        Vector3(2, 5, 6),
        Vector3(1, 8, 1),
        Vector3(10, -5, 100)
    ]

    viewport_points = [
        Vector2(10, 5),
        Vector2(100, 25)
    ]

    world_plane_points = [
        Vector3(0, 0, -10),
        Vector3(10, 0, -10),
        Vector3(0, 10, -10)
    ]

    game_area = Rectangle(0, 0, 1000, 1000)

    translation_vector = Vector3(0, 0, 0)
    rotation_point = Vector3(0, 0, 0)
    rotation_vector = None
    rotation_rads = None
    scale_vector = None
    near_plane_dist = 1
    far_plane_dist = 5
    field_of_view_rads = math.pi / 2
    matrix_state = DrawMatrixState()

    camera_orientation_args = ObjectOrientationArgs(
        translation_vector=translation_vector,
        rotation_point=rotation_point,
        rotation_axis=rotation_vector,
        rotation_rads=rotation_rads,
        scale_vector=scale_vector,
    )

    rendered_view_args = RenderedViewArgs(
        near_plane_dist=near_plane_dist,
        far_plane_dist=far_plane_dist,
        field_of_view_rads=field_of_view_rads,
        camera_orientation_args=camera_orientation_args
    )
    matrix_state.set_matricies(rendered_view_args=rendered_view_args)
    pc = MultiPerspectivePointConverter()



    pc.convert_real_to_viewport(real_points=real_points,
                                matrix_state=matrix_state,
                                game_area_rect=game_area)
    pprint(pc.real_world_points)
    pprint(pc.camera_view_world_points)
    pprint(pc.projection_points)
    pprint(pc.viewport_points)

    pc.convert_viewport_to_real(viewport_points=viewport_points,
                                game_area_rect=game_area,
                                world_plane_points=world_plane_points,
                                matrix_state=matrix_state)

    pprint(pc.real_world_points)
    pprint(pc.camera_view_world_points)
    pprint(pc.viewport_points)


import logging
from examples.voroni_world.models import *
import pprint
from typing import Dict, Tuple, List, Callable
from cooptools import graphs as grph
from cooptools.physics import waypointFollower as wf

logger = logging.getLogger(__name__)

def send_villager_to_region(
        villager: Villager,
        region: Region,
        villager_current_region_provider: Callable[[Villager], Region],
        route_provider: Callable[[Region, Region], grph.AStarResults],
        route_to_waypoints_provider: Callable[[List[grph.Node]], List[wf.Waypoint]],
        stop_at_border: bool = False
):
    current_region = villager_current_region_provider(villager)
    route = route_provider(current_region, region)

    if route.nodes is None:
        logger.error(f"Unable to route from {current_region.name} to selected region {region.name}"
                     f"\n[{pprint.pformat(route)}]")
        return False


    if stop_at_border:
        nds = list(route.nodes[:-1])
    else:
        nds = list(route.nodes)

    wpts = route_to_waypoints_provider(nds)

    villager.waypoint_follower.update(delta_time_ms=0,
                                      new_waypoints=wpts)


from coopstructs.geometry.vectors.vectorN import Vector2
from coopstructs.geometry.curves.curveBuilder import CurveBuilder
from coopstructs.geometry.curves.curve_factory import curves_from_json
from cooptools.toggles import BooleanToggleable
from coopstructs.geometry import Rectangle
from coopgame.gameTemplate import GameTemplate
import pygame
from cooptools.colors import Color
from uuid import uuid4
import coopgame.renderedObjectHandling.MatrixManipulation as mm
import numpy as np
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
import cooptools.os_manip as osm
from coopui.cli.CliAtomicUserInteraction import CliAtomicUserInteraction as cli
from coopgame.linedrawing.pyCurveHandler import PyCurveHandler

class CurveDrawGame(GameTemplate):

    def __init__(self):
        self.curve_builder = CurveBuilder(id_provider=self.get_next_id)
        input_state_handler = InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.visibility_toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyMouse.LEFT, event_type=InputEventType.DOWN)], callback=lambda x: self.handle_left_mouse_down(x)),
                InputAction([InputEvent(PyMouse.RIGHT, event_type=InputEventType.DOWN)], callback=lambda x: self.handle_right_mouse_down(x)),
                InputAction([InputEvent(PyKeys.K_s, event_type=InputEventType.DOWN)], callback=lambda x: self.save_curve(x)),
                InputAction([InputEvent(PyKeys.K_SPACE, event_type=InputEventType.DOWN)], callback=lambda x: self.toggle_curve_type(x)),
                InputAction([InputEvent(PyKeys.K_q, event_type=InputEventType.DOWN)], callback=lambda x: self.use_scaler_matrix.toggle()),
                InputAction([InputEvent(PyKeys.K_t, event_type=InputEventType.DOWN)], callback=lambda x: self.curve_builder.toggle_curve_type()),
                InputAction([InputEvent(PyKeys.K_l, event_type=InputEventType.DOWN)], callback=lambda x: self.load_curves()),
        ])


        super().__init__(fullscreen=False, input_state_handler=input_state_handler)



        self.curve_drawer = PyCurveHandler()

        self.use_scaler_matrix = BooleanToggleable(default=False)

        self.x_rot = 1.1
        self.time_since_report_stats = 0

        self.force_continuity = True

    def toggle_curve_type(self, pressed_keys):
        self.curve_builder.toggle_curve_type()

    def save_curve(self, pressed_keys):
        self.curve_builder.save_curve()

    def get_next_id(self):
        return str(uuid4())

    def initialize_game(self):
        pass

    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        pass

    def handle_left_mouse_down(self, pressed_keys):
        self.curve_builder.add_point(self.next_point())

    def handle_right_mouse_down(self, pressed_keys):
        self.curve_builder.remove_point()

    def increase_x_rotation(self):
        self.x_rot += .1

    def decrease_x_rotation(self):
        self.x_rot -= .1

    def draw(self, frames:int, debug_mode: bool = False):
        # Get saved curves
        curves = {curve: Color.GREEN for id, curve in self.curve_builder.curves.items()}
        next_point = self.next_point()

        try:
            temp_curves = self.curve_builder.curves_from_temp_next_points([next_point])

            # Append proposed curves
            for curve in temp_curves:
                curves[curve] = Color.LIGHT_BLUE
        except:
            pass

        # Draw control points
        for point in self.curve_builder.current_points:
            pygame.draw.circle(self.screen, Color.AQUA.value,
                               (int(point.x), int(point.y)), 4, 0)

        # Draw curves
        self.curve_drawer.draw_curves(curves, self.screen, draw_scale_matrix=self.draw_scale_matrix())

        # Draw next point
        game_area_rect = Rectangle.from_tuple((0, 0, self.screen.get_height(), self.screen.get_width()))
        next_point_on_plane = help.normal_points_to_scaled_points([next_point], draw_scale_matrix=self.draw_scale_matrix())[0]

        pygame.draw.circle(self.screen, Color.GREY.value,
                           (int(next_point_on_plane[0]), int(next_point_on_plane[1])), 4, 1)


    def model_updater(self, delta_time_ms: int):
        pass

    def sprite_updater(self, delta_time_ms: int):
        pass

    def on_resize(self):
        pass

    def next_point(self):
        game_area_rect = Rectangle.from_tuple((0, 0, self.screen.get_height(), self.screen.get_width()))
        if self.force_continuity:
            return self.curve_builder.force_continuity(help.normal_mouse_position(game_area_rect = game_area_rect, draw_scale_matrix=self.draw_scale_matrix()))
        else:
            return help.normal_mouse_position(draw_scale_matrix=self.draw_scale_matrix(), game_area_rect=game_area_rect)

    def draw_scale_matrix(self):
        if self.use_scaler_matrix.value:
            return mm.get_grid_rotation_matrix(self.screen_width/2, self.screen_height / 2,  self.x_rot, np.pi/4)
        else:
            return None

    def load_curves(self):
        fp = r'C:\Users\tburns\PycharmProjects\coopstructs\tests\testData\curves.json'

        if fp is None:
            fp=cli.request_open_filepath(prompt=f"Select a curves file", title="Curves", filetypes=((osm.FileType.JSON.value, ".json"), ))

        if fp is None:
            return

        data = osm.try_load_json_data_to_dict(file_path=fp)
        if data is None:
            return

        self.curve_builder.curves.clear()
        self.curve_builder.add_curves(curves_from_json(data))
        self.curve_builder.save_curve()


if __name__ == "__main__":
    import logging
    import loggingConfig

    loggingConfig.initLogging(loggingLvl=logging.INFO)
    game = CurveDrawGame()
    game.main()





from dataclasses import dataclass
from typing import Dict, Tuple, List, Callable, Any
import pygame
import pygame.event

from coopgame.pygame_k_constant_names import PyKeys, PyMouse, InputType
from cooptools.decor import try_handler
from coopgame.logger import logger
from cooptools.coopEnum import CoopEnum, auto

class InputEventType(CoopEnum):
    DOWN = auto()
    UP = auto()
    HELD = auto()
    HOVER = auto()

SupportedInput = PyKeys | PyMouse

@dataclass(frozen=True, slots=True, eq=True)
class InputEvent:
    input_key: SupportedInput
    event_type: InputEventType = InputEventType.DOWN

    @property
    def InputType(self):
        if type(self.input_key) == PyKeys:
            return InputType.KEYBOARD
        elif type(self.input_key) == PyMouse:
            return InputType.MOUSE
        else:
            raise TypeError(f"Unhandled input type {self.input_key}")

    def with_event_type(self, event_type: InputEventType):
        return InputEvent(
            input_key=self.input_key,
            event_type=event_type
        )

class InputState:
    def __init__(self,
                 delta_ms: int,
                 events: List[InputEvent] = None
                ):
        # init
        self._index: int = -1
        self._delta_ms: int = delta_ms
        self._events: Dict[int, InputEvent] = {}
        self._mouse_pos = None

        # register events that have been provided
        if events is not None:
            self.register(events)

        # Get currently pressed keys
        keys = pygame.key.get_pressed()
        self.register(events=
                      [InputEvent(event_type=InputEventType.HELD,
                                  input_key=x) for x in PyKeys if keys[x.value]]
        )

        # Get currently pressed Mouse
        self.register(events=
                      [InputEvent(event_type=InputEventType.HELD,
                                  input_key=PyMouse.by_val(ii+1)) for ii, x in enumerate(pygame.mouse.get_pressed()) if x is True]
        )

        # Register Mouse Pos Event
        self.register(events=[
            InputEvent(event_type=InputEventType.HOVER,
                       input_key=PyMouse.POSITION)
        ])

    def _get_next_index(self):
        self._index += 1
        return self._index

    def __iter__(self):
        for x, v in self._events.items():
            yield v

    def register(self,
                 events: List[InputEvent] =None,
                 mouse_pos: Tuple[float, float] = None):
        if events is not None:
            # delete any HELD (dont want both a MU and MH event in same iteration)
            for event in events:
                held_ = event.with_event_type(event_type=InputEventType.HELD)
                if held_ in self._events.values():
                    self._events = {k: v for k, v in self._events.items() if v != held_}

            # capture events
            for event in events:
                self._events[self._get_next_index()] = event

        if mouse_pos is not None:
            self._mouse_pos = mouse_pos

    @property
    def KeyDownEvents(self) -> List[PyKeys]:
        return [x.input_key for x in self._events.values() if x.InputType == InputType.KEYBOARD and x.event_type == InputEventType.DOWN]

    @property
    def KeyUpEvents(self) -> List[PyKeys]:
        return [x.input_key for x in self._events.values() if x.InputType == InputType.KEYBOARD and x.event_type == InputEventType.UP]

    @property
    def MouseDownEvents(self) -> List[PyMouse]:
        return [x.input_key for x in self._events.values() if x.InputType == InputType.MOUSE and x.event_type == InputEventType.DOWN]

    @property
    def MouseUpEvents(self) -> List[PyMouse]:
        return [x.input_key for x in self._events.values() if x.InputType == InputType.MOUSE and x.event_type == InputEventType.UP]

    @property
    def MouseHeld(self) -> List[PyMouse]:
        return [x.input_key for x in self._events.values() if x.InputType == InputType.MOUSE and x.event_type == InputEventType.HELD]

    @property
    def KeysHeld(self) -> List[PyMouse]:
        return [x.input_key for x in self._events.values() if x.InputType == InputType.KEYBOARD and x.event_type == InputEventType.HELD]

    @property
    def Events(self) -> List[InputEvent]:
        return list(self._events.values())

    @property
    def DeltaEvents(self) -> List[InputEvent]:
        return [x for x in self._events.values() if x.event_type in [InputEventType.UP, InputEventType.DOWN]]

    @property
    def DeltaMs(self) -> int:
        return self._delta_ms

    @property
    def MousePos(self) -> Tuple[float, float]:
        return self._mouse_pos

    def __repr__(self):
        return f"deltaMs: {self._delta_ms}, MU: {self.MouseUpEvents}, MD: {self.MouseDownEvents}, KU: {self.KeyUpEvents}, KD:{self.KeyDownEvents}, MH: {self.MouseHeld}, KH:{self.KeysHeld}"

def pygame_event_handler(inp_state: InputState, event: pygame.event.Event):
    if event.type == pygame.KEYDOWN:
        inp_state.register(events=[InputEvent(event_type=InputEventType.DOWN, input_key=PyKeys.by_val(event.key))])
    elif event.type == pygame.MOUSEBUTTONDOWN:
        inp_state.register(events=[InputEvent(event_type=InputEventType.DOWN, input_key=PyMouse.by_val(event.button))])
    elif event.type == pygame.KEYUP:
        inp_state.register(events=[InputEvent(event_type=InputEventType.UP, input_key=PyKeys.by_val(event.key))])
    elif event.type == pygame.MOUSEBUTTONUP:
        inp_state.register(events=[InputEvent(event_type=InputEventType.UP, input_key=PyMouse.by_val(event.button))])

    return inp_state

InputCallback = Callable[[InputState], Any]


@dataclass(frozen=True, slots=True, eq=True)
class InputAction:
    input_events: List[InputEvent]
    callback: InputCallback
    description: str = None

@dataclass(frozen=True)
class CallbackPackage:
    down: InputCallback = None
    up: InputCallback = None
    held: InputCallback = None
    hover: InputCallback = None

    def actions(self, keys: List[SupportedInput]) -> List[InputAction]:
        ret = []

        for key in keys:
            if self.down is not None:
                ret.append(InputAction(input_events=[InputEvent(input_key=key,
                                                                event_type=InputEventType.DOWN)],
                                       callback=self.down))

            if self.up is not None:
                ret.append(InputAction(input_events=[InputEvent(input_key=key,
                                                                event_type=InputEventType.UP)],
                                       callback=self.up))
            if self.held is not None:
                ret.append(InputAction(input_events=[InputEvent(input_key=key,
                                                                event_type=InputEventType.HELD)],
                                       callback=self.held))

            if self.hover is not None:
                ret.append(InputAction(input_events=[InputEvent(input_key=key,
                                                                event_type=InputEventType.HOVER)],
                                       callback=self.hover))

        return ret

class InputStateHandler:
    def __init__(self,
                 key_actions: List[InputAction] = None,
                 quit_callback_package: CallbackPackage = None,
                 debug_callback_package: CallbackPackage = None,
                 fullscreen_callback_package: CallbackPackage = None
                 ):
        self._input_actions: Dict[List[Tuple[SupportedInput]], InputAction] = {}

        to_register = key_actions if key_actions is not None else []
        if quit_callback_package is not None:
            to_register += quit_callback_package.actions([PyKeys.K_ESCAPE])

        if debug_callback_package is not None:
            to_register += debug_callback_package.actions([PyKeys.K_F12])

        if fullscreen_callback_package is not None:
            to_register += fullscreen_callback_package.actions([PyKeys.K_F11])

        self.register_key_actions(to_register)

    @property
    def InputActions(self):
        return self._input_actions

    def register_key_actions(self, key_actions: List[InputAction]):
        self._input_actions = {**self._input_actions, **{tuple(x.input_events): x for x in key_actions}}

    @staticmethod
    def _handle_callback(input_action, input_state):
        logger.info(f"Handling callback for input_action {input_action} with state {input_state}")
        input_action.callback(input_state)


    @try_handler(logger=logger)
    def handle_input(self, input_state: InputState):

        # log state
        if len(input_state.Events) > 0:
            logger.debug(input_state)

        if len(input_state.DeltaEvents) > 0:
            logger.info(f"Pressed Keys: [{','.join([x.input_key.name for x in input_state.DeltaEvents if x.event_type==InputEventType.DOWN])}] "
                        f"Released Keys: [{','.join([x.input_key.name for x in input_state.DeltaEvents if x.event_type==InputEventType.UP])}]")

        # handle input
        for input_action in self._input_actions.values():
            # if, by chance a trigger button was pressed twice in one input_state, make sure to handle the callback twice
            triggers = [x for x in input_action.input_events if x.event_type in [InputEventType.UP, InputEventType.DOWN]]
            trigger_count = len(triggers)
            n_times = trigger_count if trigger_count > 1 else 1

            # check to make sure all events have occurred
            if all(x in input_state.Events for x in input_action.input_events):
                for n in range(n_times):
                    self._handle_callback(input_action=input_action, input_state=input_state)

        # # handle hover over
        # if input_state.MousePos is not None and self._hover_callback is not None:
        #     MonitoredClassTimer.timer(self._hover_callback(input_state))

    def with_debug(self,
                   debug_callback: Callable,
                   key: PyKeys = PyKeys.K_F12):
        cbp = CallbackPackage(down=debug_callback)
        self.register_key_actions(cbp.actions([key]))
        return self

    def with_number_mapping(self,
                            number_mappings: Dict[int, Callable]):
        num_lookup = {
            0: [PyKeys.K_0, PyKeys.K_KP0],
            1: [PyKeys.K_1, PyKeys.K_KP1],
            2: [PyKeys.K_2, PyKeys.K_KP2],
            3: [PyKeys.K_3, PyKeys.K_KP3],
            4: [PyKeys.K_4, PyKeys.K_KP4],
            5: [PyKeys.K_5, PyKeys.K_KP5],
            6: [PyKeys.K_6, PyKeys.K_KP6],
            7: [PyKeys.K_7, PyKeys.K_KP7],
            8: [PyKeys.K_8, PyKeys.K_KP8],
            9: [PyKeys.K_9, PyKeys.K_KP9],
        }

        for i, c in number_mappings.items():
            if not i in num_lookup.keys():
                raise ValueError(f"{i} not a valid number mapping input")
            keys = num_lookup[i]

            acts = []
            for key in keys:
                ia = InputAction([InputEvent(key, event_type=InputEventType.DOWN)],
                         callback=c)
                acts.append(ia)

            self.register_key_actions(acts)

        return self

    def with_directionals(self,
                          up_callback: Callable,
                          down_callback: Callable,
                          left_callback: Callable,
                          right_callback: Callable,
                          while_holding: bool = False):
        direction_map = {
            left_callback: [PyKeys.K_LEFT, PyKeys.K_a, PyKeys.K_KP4],
            right_callback: [PyKeys.K_RIGHT, PyKeys.K_d, PyKeys.K_KP6],
            up_callback: [PyKeys.K_UP, PyKeys.K_w, PyKeys.K_KP8],
            down_callback: [PyKeys.K_DOWN, PyKeys.K_s, PyKeys.K_KP5],
        }

        for cb, keys in direction_map.items():
            acts = []
            for key in keys:
                if while_holding:
                    et = InputEventType.HELD
                else:
                    et = InputEventType.DOWN

                ia = InputAction([InputEvent(key, event_type=et)], callback=cb)
                acts.append(ia)

            self.register_key_actions(acts)

        return self

    def with_pause(self,
                   pause_callback: Callable,
                   key: PyKeys = PyKeys.K_p
                   ):
        cbp = CallbackPackage(down=pause_callback)
        self.register_key_actions(cbp.actions([key]))
        return self

def basic_key_actions(
        up_package: CallbackPackage = None,
        down_package: CallbackPackage = None,
        left_package: CallbackPackage = None,
        right_package: CallbackPackage = None,
        rot_left_package: CallbackPackage = None,
        rot_right_package: CallbackPackage = None,
        scale_up_package: CallbackPackage = None,
        scale_down_package: CallbackPackage = None,
        quit_package: CallbackPackage = None,
        fullscreen_package: CallbackPackage = None,
        debug_mode_package: CallbackPackage = None,
        pause_package: CallbackPackage = None,
        runspeed_package: (CallbackPackage, CallbackPackage) = None,
        help_package: CallbackPackage = None
) -> List[InputAction]:
    ret: List[InputAction] = []

    if up_package:
        ret += up_package.actions(keys=[PyKeys.K_UP, PyKeys.K_w, PyKeys.K_KP8])

    if down_package:
        ret += down_package.actions(keys=[PyKeys.K_DOWN, PyKeys.K_s, PyKeys.K_KP2])

    if left_package:
        ret += left_package.actions(keys=[PyKeys.K_LEFT, PyKeys.K_a, PyKeys.K_KP4])

    if right_package:
        ret += right_package.actions(keys=[PyKeys.K_RIGHT, PyKeys.K_d, PyKeys.K_KP6])

    if rot_left_package:
        ret += rot_left_package.actions(keys=[PyKeys.K_q, PyKeys.K_COMMA, PyKeys.K_PAGEUP])

    if rot_right_package:
        ret += rot_right_package.actions(keys=[PyKeys.K_e, PyKeys.K_PERIOD, PyKeys.K_PAGEDOWN])

    if scale_up_package:
        ret += scale_up_package.actions(keys=[PyKeys.K_PLUS, PyKeys.K_EQUALS, PyKeys.SCROLLUP, PyKeys.K_KP_PLUS])

    if scale_down_package:
        ret += scale_down_package.actions(keys=[PyKeys.K_MINUS, PyKeys.SCROLLDOWN, PyKeys.K_KP_MINUS])

    if quit_package:
        ret += quit_package.actions(keys=[PyKeys.K_ESCAPE])

    if fullscreen_package:
        ret += fullscreen_package.actions(keys=[PyKeys.K_F12])

    if debug_mode_package:
        ret += debug_mode_package.actions(keys=[PyKeys.K_F11])

    if pause_package:
        ret += pause_package.actions(keys=[PyKeys.K_p])

    if runspeed_package:
        ret += runspeed_package[0].actions(keys=[PyKeys.K_PLUS, PyKeys.K_EQUALS])
        ret += runspeed_package[1].actions(keys=[PyKeys.K_MINUS])

    if help_package:
        ret += help_package.actions(keys=[PyKeys.K_h])

    return ret





from coopstructs.geometry.vectors.vectorN import Vector2
from coopgame.particleSystem.callables import float_from_ms_callable
from typing import Union, Tuple
from cooptools.colors import Color
import uuid
from coopgame.particleSystem.callables import vec2_callable

class Particle:

    def __init__(self,
                 pos: Vector2,
                 velo_u_s: Union[Vector2, vec2_callable],
                 accel_u_s2: Union[Vector2, vec2_callable],
                 size: Union[float, float_from_ms_callable],
                 context = None,
                 id = None):

        if pos is None:
            raise ValueError(f"Particle pos cannot be initiated as None")

        self.pos = pos
        self.velo_u_s = velo_u_s
        self.accel_u_s2 = accel_u_s2 or Vector2(0, 0)
        self._lifetime_ms = 0
        self._size: Union[float, float_from_ms_callable] = size

        self._size_resolved: float = None
        self._color_resolved: Tuple[int, int, int] = None

        self.update(0)
        self.context = context
        self.id = id or uuid.uuid4()

    def __hash__(self):
        return hash(self.id)

    def _resolve_size(self, size, lifetime_ms):
        if callable(size):
            return size(lifetime_ms)

        return size

    def _resolve_velo(self):
        if callable(self.velo_u_s):
            return self.velo_u_s()
        return self.velo_u_s

    def _resolve_accel(self):
        if callable(self.accel_u_s2):
            return self.accel_u_s2()
        return self.accel_u_s2

    def update(self, delta_time_ms: int):
        self._lifetime_ms += delta_time_ms
        self._size_resolved = self._resolve_size(self._size, self._lifetime_ms)

        self.pos += self._resolve_velo() * delta_time_ms / 1000
        self.velo_u_s += self._resolve_accel() * delta_time_ms / 1000

    @property
    def Lifetime_ms(self):
        return self._lifetime_ms

    @property
    def Size(self):
        return self._size_resolved

    @property
    def Color(self):
        return Color.closest_color(self._color_resolved)

    @property
    def Color_RGB(self):
        return self._color_resolved



from coopstructs.geometry.vectors.vectorN import Vector2
import math
import random as rnd
from dataclasses import dataclass


def rotation(time_ms, r_high, r_wide, phase: int = 0) -> Vector2:
    return Vector2(((math.sin(time_ms + phase) + 1) * r_wide) / 2, ((math.cos(time_ms + phase) + 1) * r_high) / 2)

def rand_between_points(a: Vector2, b: Vector2) -> Vector2:
    return a.interpolate(b, rnd.random())

@dataclass(frozen=True)
class SinWaveArgs:
    amp: float = 1
    wavelength: float = 1
    phase: float = 0
    norm_positive: bool = False

def sin_wave(x: float, args: SinWaveArgs):
    sin_val = math.sin((x + args.phase) / args.wavelength)
    if args.norm_positive:
        sin_val = (1 + sin_val) / 2

    return sin_val * args.amp

def decay(x: float, lifespan_ms: float, max_size: float, end_size: float = 0):
    _r = 1 / lifespan_ms
    return end_size + (1 - _r * x) * (max_size - end_size)

def grow(x: float, lifespan_ms: float, max_size: float, start_size: float = 0):
    _r = 1 / lifespan_ms
    return start_size + (_r * x) * (max_size - start_size)

def pulse(x: float, lifespan_ms: float, max_size: float, min_size: float, period_ms: float):
    pass



from cooptools.colors import Color
import pygame
from cooptools.graphs import Graph, Edge, Node, AStarResults
from coopstructs.geometry.vectors.vectorN import Vector2
from coopstructs.geometry import Rectangle
import coopgame.pygamehelpers as help
from typing import Callable, Tuple, Dict, Iterable
from coopstructs.geometry.curves.curves import LineCurve, Curve
from cooptools.coopEnum import Directionality
from dataclasses import dataclass, asdict
import numpy as np
from coopgame.pointdrawing import point_draw_utils as putils
from coopgame.label_drawing import label_drawing_utils as lutils
from coopgame.linedrawing import line_draw_utils as nutils
from coopstructs.geometry.lines import Line
import cooptools.matrixManipulation as mm
import cooptools.geometry_utils.vector_utils as vec
from cooptools.common import flattened_list_of_lists
from coopgame.surfaceManager import SurfaceGroup

@dataclass(frozen=True)
class RouteDrawArgs:
    edge_draw_args: nutils.DrawLineArgs = None
    node_draw_args: putils.DrawPointArgs = None
    node_label_args: lutils.DrawLabelArgs = None

@dataclass(frozen=True)
class GraphDrawArgs:
    node_draw_args: putils.DrawPointArgs = None
    enabled_edge_args: nutils.DrawLineArgs = None
    disabled_edge_args: nutils.DrawLineArgs = None
    node_label_args: lutils.DrawLabelArgs = None
    articulation_points_args: putils.DrawPointArgs = None
    source_node_args: putils.DrawPointArgs = None
    sink_node_args: putils.DrawPointArgs = None
    orphan_node_args: putils.DrawPointArgs = None

    def with_(self, **kwargs):
        definition = self.__dict__
        for kwarg, val in kwargs.items():
            definition[kwarg] = val

        return type(self)(**definition)

    @classmethod
    def from_(cls, args, **kwargs):
        kw = asdict(args)
        kw.update(kwargs)
        return GraphDrawArgs(**kw)

    @property
    def EdgesBaseArgs(self):
        return GraphDrawArgs(
            enabled_edge_args=self.enabled_edge_args.BaseArgs if self.enabled_edge_args is not None else None,
            disabled_edge_args=self.disabled_edge_args.BaseArgs if self.disabled_edge_args is not None else None
        )

    @property
    def EdgesLabelArgs(self):
        return GraphDrawArgs(
            enabled_edge_args=self.enabled_edge_args,
            disabled_edge_args=self.disabled_edge_args
        )

    @property
    def NodesBaseArgs(self):
        return GraphDrawArgs(
            node_draw_args=self.node_draw_args
        )

    @property
    def NodesLabelArgs(self):
        return GraphDrawArgs(
            node_label_args=self.node_label_args
        )

    @property
    def OverlayArgs(self):
        return GraphDrawArgs(
            articulation_points_args=self.articulation_points_args,
            source_node_args=self.source_node_args,
            sink_node_args=self.sink_node_args,
            orphan_node_args=self.orphan_node_args
        )


def get_node_label_surface_group(
    graph: Graph,
    args: GraphDrawArgs,
    vec_transformer: vec.VecTransformer = None
) -> SurfaceGroup:
    positions = {x.name: vec.resolve_vec_transformed_points([x.pos], vec_transformer)[0][:2] for x in graph.Nodes.values()}

    return lutils.label_surface_group(
        labels={x.name: x.name for x in graph.Nodes.values()},
        draw_label_args=args.node_label_args,
        positions=positions
    )

def get_edge_label_surface_group(
    graph: Graph,
    args: GraphDrawArgs,
    vec_transformer: vec.VecTransformer = None
) -> SurfaceGroup:
    positions = {x.id: vec.resolve_vec_transformed_points(points=[vec.interpolate(x.start.pos, x.end.pos)],
                                                          vec_transformer=vec_transformer)[0][:2]
                 for x in graph.Edges.values()}

    return lutils.label_surface_group(
        labels={x.id: str(x) for x in graph.Edges.values()},
        draw_label_args={x.id: args.enabled_edge_args.draw_label_args
        if x.enabled else args.disabled_edge_args.draw_label_args
                         for x in graph.Edges.values()},
        positions=positions
    )

def draw_to_surface(surface: pygame.Surface,
                    graph: Graph,
                    args: GraphDrawArgs = None,
                    routes: Iterable[Tuple[AStarResults, RouteDrawArgs]] = None,
                    vec_transformer: vec.VecTransformer = None):
    if graph is None:
        return

    if args is not None and (args.enabled_edge_args or args.disabled_edge_args):
        _draw_edges(
            surface=surface,
            edge_draw_args={
                e: args.enabled_edge_args if e.enabled() else args.disabled_edge_args for e in graph.Edges.values()
            },
            vec_transformer=vec_transformer
        )

    if args is not None and (args.node_draw_args or args.node_label_args):
        draw_graph_nodes(surface,
                         nodes=graph.Nodes.values(),
                         node_draw_args=args.node_draw_args,
                         vec_transformer=vec_transformer,
                         draw_label_args=args.node_label_args)

    if routes is not None:
        draw_routes(
            surf=surface,
            routes=routes,
            vec_transformer=vec_transformer
        )


    if args is not None and args.articulation_points_args:
        putils.draw_points(
            points={x.pos: args.articulation_points_args for x in graph.ArticulationPoints},
            surface=surface,
            vec_transformer=vec_transformer
        )

    if args is not None and args.source_node_args:
        putils.draw_points(
            points={x.pos: args.source_node_args for x in graph.Sources},
            surface=surface,
            vec_transformer=vec_transformer
        )

    if args is not None and args.sink_node_args:
        putils.draw_points(
            points={x.pos: args.sink_node_args for x in graph.Sinks},
            surface=surface,
            vec_transformer=vec_transformer
        )

    if args is not None and args.orphan_node_args:
        putils.draw_points(
            points={x.pos: args.orphan_node_args for x in graph.Orphans},
            surface=surface,
            vec_transformer=vec_transformer
        )
def _draw_edges(
    surface: pygame.Surface,
    edge_draw_args: Dict[Edge, nutils.DrawLineArgs],
    vec_transformer: vec.VecTransformer = None
):
    line_args = {Line(Vector2.from_tuple(k.start.pos),
                      Vector2.from_tuple(k.end.pos)): v.with_(label_text=str(k))
                 for k, v in edge_draw_args.items() if k.start.pos != k.end.pos}

    nutils.draw_lines(lines=line_args,
                      surface=surface,
                      vec_transformer=vec_transformer
                      )

def draw_graph_nodes(surface: pygame.Surface,
                     nodes: Iterable[Node],
                     node_draw_args: putils.DrawPointArgs = None,
                     draw_label_args: lutils.DrawLabelArgs = None,
                     vec_transformer: vec.VecTransformer = None):
    txt_dict = {}
    for node in nodes:
        if node_draw_args:
            putils.draw_points({node.pos: node_draw_args},
                               surface=surface,
                               vec_transformer=vec_transformer)


        scaled = vec.resolve_vec_transformed_points([node.pos], vec_transformer=vec_transformer)[0]
        txt_dict.setdefault(scaled, []).append(node.name)


    if draw_label_args:
        for pos, labels in txt_dict.items():
            lutils.draw_label(
                hud=surface,
                args=lutils.LabelArgs(
                    text=",".join(labels),
                    draw_args=draw_label_args,
                    pos=pos
                )
            )

def draw_articulation_points(surface: pygame.Surface,
                             graph: Graph,
                             color: Color = None,
                             vec_transformer: vec.VecTransformer = None
                             ):
    if graph is None:
        return

    articulation_points = graph.AP()

    if color is None:
        color = Color.ORANGE


    scaled_pos = vec.resolve_vec_transformed_points(points=[node.pos for node in articulation_points.keys()],
                                                    vec_transformer=vec_transformer)

    for point in scaled_pos:
        pygame.draw.circle(surface, color.value, (int(point[0]), int(point[1])), 10, 1)

def draw_directionality_indicators(curves: Dict[Curve, Directionality],
                                   surface: pygame.Surface,
                                   indicator_color: Color,
                                   num_arrows: int = 5,
                                   size: float = 1,
                                   indicator_points_color: Color | Tuple[Color, ...]=None,
                                   vec_transformer: vec.VecTransformer = None):
    arrow_ts = [1.0 / (num_arrows - 1) * x for x in range(0, num_arrows)] if num_arrows > 1 else [0.5]


    for curve, direction in curves.items():
        for t in arrow_ts:
            centre = curve.point_at_t(t)

            try:
                # get derivative of curve for drawing
                derivative = curve.derivative_at_t(t)
                d_unit = derivative.unit()
            except:
                # most likely a vertical curve (no derivative), handle by pointing up or down.
                d_unit = (curve.EndPoint - curve.origin).unit()

            if d_unit is None or d_unit.y == 0:
                continue

            d_foreward = d_unit.scaled_to_length(size)
            d_ort_1 = Vector2(1, - d_unit.x / d_unit.y).scaled_to_length(size / 2)
            # d_ort_2 = d_ort_1 * -1

            a = b = c = d = e = f = None
            if direction in [Directionality.FOREWARD, Directionality.BIDIRECTIONAL]:
                a = centre
                b = centre - d_foreward + d_ort_1
                c = centre - d_foreward - d_ort_1

                scaled_polygon_points = vec.resolve_vec_transformed_points(points=[a.as_tuple(), b.as_tuple(), c.as_tuple()],
                                                                           vec_transformer=vec_transformer)

                help.draw_polygon(surface, scaled_polygon_points, color=indicator_color)

            if direction in [Directionality.BACKWARD, Directionality.BIDIRECTIONAL]:
                d = centre
                e = centre + d_foreward + d_ort_1
                f = centre + d_foreward - d_ort_1

                scaled_polygon_points = vec.resolve_vec_transformed_points(points=[d.as_tuple(), e.as_tuple(), f.as_tuple()],
                                                                           vec_transformer=vec_transformer)

                help.draw_polygon(surface, scaled_polygon_points, color=indicator_color)

            ip_color_getter = lambda ii: indicator_points_color[ii % len(indicator_points_color)] \
                if type(indicator_points_color) == tuple \
                else indicator_points_color

            if indicator_points_color:
                points = [a, b, c, d, e, f]
                putils.draw_points(
                    points={
                        x.as_tuple(): putils.DrawPointArgs(
                            color=ip_color_getter,
                            outline_color=None,
                            radius=2
                        ) for ii, x in enumerate(points)
                    },
                    surface=surface,
                    vec_transformer=vec_transformer
                )

def draw_routes(
        surf: pygame.Surface,
        routes: Iterable[Tuple[AStarResults, RouteDrawArgs]],
        vec_transformer: vec.VecTransformer = None
):
    if routes is None:
        return

    for route, draw_args in routes:
        if route.nodes is None:
            continue

        _draw_edges(surface=surf,
                    edge_draw_args={e: draw_args.edge_draw_args for e in route.edges},
                    vec_transformer=vec_transformer)
        draw_graph_nodes(
            surface=surf,
            nodes=route.nodes,
            node_draw_args=draw_args.node_draw_args,
            draw_label_args=draw_args.node_label_args,
            vec_transformer=vec_transformer
        )


from cooptools.colors import Color
import harfang as hf

def harfang_color_from_coopcolor(coopcolor: Color = None):
    if coopcolor is None:
        coopcolor = Color.Blue

    ctup = coopcolor.value
    return hf.Color(ctup[0], ctup[1], ctup[2])


import pygame
from coopstructs.geometry.vectors.vectorN import Vector2
from coopstructs.toggles import BooleanToggleable
from coopgame.gameTemplate import GameTemplate
from coopgraph.graphs import Graph, Node
from cooptools.sectors.grids import RectGrid
from cooptools.dictPolicies import TogglePolicy
from coopgame.grids_graphs.pygridhandler import PyGridHandler, GridDrawType
from coopgame.grids_graphs.pygraphhandler import PyGraphHandler
from cooptools.toggles import BooleanToggleable
from cooptools.colors import Color
import coopgame.pygamehelpers as help
from coopgame.linedrawing.pyCurveHandler import CurveDrawer
from coopgame.linedrawing.unitpathhandler import UnitPathHandler
from uuid import uuid4
import coopgame.renderedObjectHandling.MatrixManipulation as mm
import numpy as np
import json
import datetime
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse

class CurveDrawGame(GameTemplate):

    def __init__(self):
        self.use_scaler_matrix = BooleanToggleable()

        input_state_handler = InputStateHandler(quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
                                                debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.visibility_toggle()),
                                                fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
                                                key_actions=[
            InputAction([InputEvent(PyMouse.LEFT, event_type=InputEventType.DOWN)], callback=lambda x: self.handle_left_mouse_down()),
            InputAction([InputEvent(PyKeys.K_s, event_type=InputEventType.DOWN)], callback=lambda x: self.curves_to_graph()),
            InputAction([InputEvent(PyKeys.K_q, event_type=InputEventType.DOWN)], callback=lambda x: self.use_scaler_matrix.toggle()),
            InputAction([InputEvent(PyKeys.K_d, event_type=InputEventType.DOWN)], callback=lambda x: self.toggle_show_toggled()),
            InputAction([InputEvent(PyKeys.K_p, event_type=InputEventType.DOWN)], callback=lambda x: self.persist_graph_to_json()),
            InputAction([InputEvent(PyKeys.K_g, event_type=InputEventType.DOWN)], callback=lambda x: self.persist_grid()),
            InputAction([InputEvent(PyKeys.K_l, event_type=InputEventType.DOWN)], callback=lambda x: self.load_grid()),
        ])

        super().__init__(fullscreen=False, input_state_handler=input_state_handler)

        self.toggle_key = 'toggle'
        self.unit_path_handler = UnitPathHandler(self.toggle_key)
        self.toggle_bool_policy = TogglePolicy(key=self.toggle_key, toggle_type=BooleanToggleable, default=False)
        self.grid = RectGrid(20, 20, default_state={})
        for pos, state in self.grid.grid_enumerator:
            self.toggle_bool_policy.initialize(state.state)

        self.grid_background_surface = None
        self.grid_overlay_surface = None
        self.curves_surface = None
        self.graph_surface = None

        self.grid_handler = PyGridHandler()
        self.graph_handler = PyGraphHandler()
        self.curves = []
        self.graph = None
        self.curve_drawer = CurveDrawer()

        self.x_rot = 0
        self.y_rot = 0
        self.time_since_report_stats = 0

        self.show_toggled = False

        self._update_grid_background_surface()


    def get_next_id(self):
        return str(uuid4())

    def _regenerate_curves(self):
        self.curves = self.unit_path_handler.generate(self.grid,
                                        self.grid_handler.grid_box_rectangle(self.screen.get_width(), self.screen.get_height(), self.grid, margin=0),
                                        naming_provider=self.get_next_id)

    def handle_left_mouse_down(self):
        grid_pos = self.mouse_grid_pos()
        self.grid.act_at_loc(grid_pos[0], grid_pos[1], [self.toggle_bool_policy])
        self._regenerate_curves()
        self._update_curves_surface()

    def on_resize(self):
        self._update_grid_background_surface()
        self._regenerate_curves()

    def mouse_grid_pos(self):
        return self.grid_handler.get_mouse_grid_pos(self.game_area_rectangle,
                                                    help.mouse_pos_as_vector(),
                                             draw_scale_matrix=self.draw_scale_matrix(),
                                             coord_to_grid_converter=self.grid.grid_from_coord)

    def toggle_show_toggled(self):
        self.show_toggled = not self.show_toggled

    def _update_grid_background_surface(self):
        self.grid_background_surface = help.init_surface(self.screen.get_size())
        self.grid_handler.draw_base_to_surface(self.grid_background_surface,
                                               self.grid, grid_color=Color.DARK_GREY,
                                               grid_draw_type=GridDrawType.LINES,
                                               draw_scale_matrix=self.draw_scale_matrix())

    def _update_grid_overlay_surface(self):
        self.grid_overlay_surface = help.init_surface(self.screen.get_size())
        self.grid_handler.draw_overlay_to_surface(surface=self.grid_overlay_surface,
                                                  grid=self.grid,
                                                  hover=(Vector2.from_tuple(self.mouse_grid_pos()), Color.PURPLE),
                                                  toggled=(self.toggle_key, Color.ORANGE) if self.show_toggled else None,
                                                  draw_scale_matrix=self.draw_scale_matrix())

    def _update_curves_surface(self):
        self.curves_surface = help.init_surface(self.screen.get_size())
        self.curve_drawer.draw_curves({x: Color.GREEN for x in self.curves}, self.curves_surface)

    def _update_graph_surface(self):
        self.graph_surface = help.init_surface(self.screen.get_size())
        if self.graph:
            self.graph_handler.draw_to_surface(self.graph_surface, self.graph)

    def draw(self, frames:int, debug_mode: bool = False):
        self._update_grid_overlay_surface()

        # draw grid base
        if self.grid_background_surface is not None:
            self.screen.blit(self.grid_background_surface, (0, 0))

        # draw grid overlay
        if self.grid_overlay_surface is not None:
            self.screen.blit(self.grid_overlay_surface, (0, 0))

        # draw unit paths
        if self.curves_surface is not None:
            self.screen.blit(self.curves_surface, (0, 0))

        #draw graph
        if self.graph_surface is not None:
            self.screen.blit(self.graph_surface, (0, 0))


    def curves_to_graph(self):

        a = np.asarray([(curve.origin, curve.end_point, type(curve)) for curve in self.curves])

        point_names = {}
        for o, d, _ in a:
            if not o in point_names:
                point_names[o] = self.get_next_id()

            if not d in point_names:
                point_names[d] = self.get_next_id()


        g = {}
        for o, d, t in a:
            n_o = Node(name=point_names[o], pos=o)
            n_d = Node(name=point_names[d], pos=d)
            g.setdefault(n_o, [])
            g.setdefault(n_d, [])

            g[n_o].append(n_d)
            g[n_d].append(n_o)

        self.graph = Graph(g)

        print(self.graph)

        self._update_graph_surface()

    def persist_graph_to_json(self):
        if self.graph is None:
            return

        my_zone = self.get_next_id()
        my_json_dict = {}

        my_json_dict["version"] = "0.0.1"
        my_json_dict["dateGenerated"] = str(datetime.datetime.utcnow().isoformat())
        my_json_dict["nodes"] = [{"nodeId": node.id,
                                  "label": "some node label",
                                  "type": "node",
                                  "locationId": ["ASDFG"],
                                  "actions": [{
                                               "zHeight":1,
                                               "action":"put",
                                               "locationId":""
                                            },
                                              {
                                                  "zHeight": 1,
                                                  "action": "pick",
                                                  "locationId": ""
                                              }
                                  ] ,
                                 "zones": [my_zone]
                                  } for node in self.graph.nodes()]
        my_json_dict["zones"] = {
            "id": my_zone,
            "zoneActions": [
                "honk"
            ],
            "enclosedNodes": [node.id for node in self.graph.nodes()],
            "polygonPoints": [{
                                       "x":1,
                                       "y":1
                                    },
                                    {
                                       "x":1,
                                       "y":1
                                    },
                                    {
                                       "x":1,
                                       "y":1
                                    },
                                    {
                                       "x":1,
                                       "y":1
                                    }],
            "metaData": {}
        }
        my_json_dict["agents"] = [
          {
             "agentId":"srt9600",
             "version":"0.0.0"
          }
       ]
        my_json_dict["graphs"] = {
            "srt9600": {
                "default":{
                    node.id: {
                        "location":{
                            "x": int(node.focal_point.x),
                            "y": int(node.focal_point.y),
                            "z": 0
                        },
                        "inHeadingDegree": 90,
                        "outHeadingDegree": 90,
                        "edges": {
                            self.graph._edges[edge_id].id:{
                                "destNode": self.graph._edges[edge_id].end.id,
                                "distEstimate": self.graph._edges[edge_id].length,
                                "blockedNodes":[],
                                "curves": [],
                                "metadta": {}
                            }
                            for edge_id in edge_ids
                        }

                    } for node, edge_ids in self.graph._node_edge_map.items()
                }
            }
        }


        my_json = json.dumps(my_json_dict, indent=4)
        print(my_json)

        with open('coop_generated_path.json', 'w') as outfile:
            outfile.write(my_json)

    def persist_grid(self):
        data_to_save = self.grid.state_value_as_array(self.toggle_key)

        my_json = json.dumps(data_to_save, indent=4)

        print(my_json)

    def load_grid(self):
        loaded_data = json.load()

    def draw_scale_matrix(self):
        if self.use_scaler_matrix.value:
            return mm.get_grid_rotation_matrix(self.screen_width/2, self.screen_height / 2,  self.x_rot, self.y_rot)
        else:
            return None

if __name__ == "__main__":
    import logging
    import loggingConfig

    loggingConfig.initLogging(loggingLvl=logging.INFO)
    game = CurveDrawGame()
    game.main()





from coopgame.gameTemplate import GameTemplate
from coopstructs.vectors import Vector2
from coopstructs.geometry import Rectangle as cRect
import pygame
from cooptools.colors import Color
from coopgame.spriteHandling.sprites import RectangleSprite, AnimatedSprite
from coopgame.grids_graphs.pygridhandler import PyGridHandler
from coopgame.grids_graphs.pygraphhandler import PyGraphHandler
from coopgraph.grids import RectGrid
from coopgraph.gridgraph import GridGraph
from coopgame.enums import GridDrawType
from coopgame.linedrawing.unitpathhandler import UnitPathHandler
import coopgame.pygamehelpers as help
from coopstructs.geometry import Rectangle
from coopgraph.gridSelectPolicies import TogglePolicy
from coopgraph.toggles import BooleanToggleable
import coopgame.renderedObjectHandling.MatrixManipulation as mm
import numpy as np
from pygame import mixer
from coopgame.spriteHandling.spritesheet import SpriteSheet
from coopgame.linedrawing.pyCurveHandler import CurveDrawer
from coopstructs.curves import LineCurve

class MyGame(GameTemplate):
    def __init__(self):
        super().__init__(fullscreen=False)

        self.agents = {}
        self.sprite_group = pygame.sprite.Group()

        self.register_action_to_keys((pygame.K_r,), self.sprite_group.empty)
        self.register_action_to_keys((pygame.K_SPACE,), self.create_yellow)
        self.register_action_to_keys((pygame.K_o,), self.set_astar_origin)
        self.register_action_to_keys((pygame.K_t,), self.toggle_show_graph)
        self.register_action_to_keys((pygame.K_g,), self.toggle_show_grid)
        self.register_action_to_keys((pygame.K_h,), self.toggle_show_grid_highlights)
        self.register_action_to_keys((pygame.K_m,), self.increase_x_rotation)
        self.register_action_to_keys((pygame.K_n,), self.decrease_x_rotation)
        self.register_action_to_keys((pygame.K_w,), lambda: self.set_character_animation("walkback"))
        self.register_action_to_keys((pygame.K_a,), lambda: self.set_character_animation("walkleft"))
        self.register_action_to_keys((pygame.K_s,), lambda: self.set_character_animation("walkdown"))
        self.register_action_to_keys((pygame.K_d,), lambda: self.set_character_animation("walkright"))
        self.register_action_to_keys((pygame.K_k,), lambda: self.set_character_animation("fall"))

        self.toggle_name = 'toggle'
        self.toggle_bool_policy = TogglePolicy(key=self.toggle_name, toggle=BooleanToggleable(default=False))
        self.grid = RectGrid(20, 20, grid_select_policies=[self.toggle_bool_policy])
        self.grid_graph = GridGraph(self.grid)
        self.grid_drawer = PyGridHandler()
        self.graph_drawer = PyGraphHandler()
        self.unit_path_drawer = UnitPathHandler(self.toggle_name)

        self.init_dynamic_surfaces()

        self.grid_foreground_surface = None
        self.grid_background_surface = None
        self.graph_surface = None

        self.curve_drawer = CurveDrawer()

        self.init_static_surfaces()
        self.init_dynamic_surfaces()

        self.x_rot = 1.1
        self._update_background_surfaces(Vector2(0, 0), Vector2(0, 0))
        self._update_foreground_surfaces(Vector2(0, 0), Vector2(0, 0))

        self._astar_curves = None
        self._astar_start_point = Vector2(0, 0)

        self.show_graph = False
        self.show_grid = True
        self.show_grid_highlights = True

        self.time_since_report_stats = 0

        self._index = 0

        # character_file = 'coopgame/assets/mage walking poses sheet copy.png'
        # self.character_spritesheet = SpriteSheet(character_file, n_rows=4, n_columns=9, x_margin=17, x_padding=34, y_margin=9, y_padding=10)
        character_file = '../coopgame/assets/spriteSheets/DA_male2.png'
        self.character_spritesheet = SpriteSheet(character_file, n_rows=21, n_columns=13, x_margin_left=17, x_padding=34, y_margin_top=15, y_padding=17, y_margin_bottom=1)

        char_pos = self.grid_drawer.grid_point_to_viewport_point(self.screen, self.grid, Vector2(0, 0), draw_scale_matrix=self.get_grid_rotation_matrix())
        char_pos = Vector2(char_pos.x, char_pos.y -256//4//2)
        self.main_character = AnimatedSprite("main"
                                             , char_pos
                                             , {"walkback": tuple(self.character_spritesheet.load_row_strip(8, colorkey=-1)[:9]),
                                                "walkleft": tuple(self.character_spritesheet.load_row_strip(9, colorkey=-1)[:9]),
                                                "walkdown": tuple(self.character_spritesheet.load_row_strip(10, colorkey=-1)[1:9]),
                                                "walkright": tuple(self.character_spritesheet.load_row_strip(11, colorkey=-1)[:9]),
                                                "fall": tuple(self.character_spritesheet.load_row_strip(20, colorkey=-1)[:6])}
                                             , default_animation_key="walkdown")

        self.animate_timer = 0
        self.animate_cycle = 100

    def initialize_game(self):
        mixer.music.load('../coopgame/assets/music/cute_bass.mp3')
        mixer.music.play(loops=-1)


    def set_character_animation(self, animation_key):
        self.main_character.set_animation(animation_key)

    def increase_x_rotation(self):
        self.x_rot += .1
        self._update_background_surfaces(Vector2(0, 0), Vector2(0, 0))
        self._update_foreground_surfaces(Vector2(0, 0), Vector2(0, 0))

    def decrease_x_rotation(self):
        self.x_rot -= .1
        self._update_background_surfaces(Vector2(0, 0), Vector2(0, 0))
        self._update_foreground_surfaces(Vector2(0, 0), Vector2(0, 0))

    def get_grid_rotation_matrix(self):
        orientation = np.array([0, 0, 1])
        rotation_unit_vector = mm.rotation_unit_vector(orientation)
        matrixz = mm.rotateAroundAxis((self.screen_width/2, self.screen_height / 2, 0), rotation_unit_vector, np.pi/4)
        orientation = np.array([1, 0, 0])
        rotation_unit_vector = mm.rotation_unit_vector(orientation)
        matrixx = mm.rotateAroundAxis((self.screen_width / 2, self.screen_height / 2, 0), rotation_unit_vector,
                                      self.x_rot)
        # matrixz = mm.rotateZMatrix(np.pi/2)
        # # matrixx = mm.rotateXMatrix(np.pi/4)
        matrix = matrixx.dot(matrixz)
        # return matrix
        return None

    def get_next_index(self):
        self._index += 1
        return self._index

    def init_dynamic_surfaces(self):
        self.grid_foreground_surface = pygame.Surface(self.screen.get_size()).convert()
        self.grid_foreground_surface.set_colorkey(Color.BLACK.value)

    def init_static_surfaces(self):
        self.grid_background_surface = pygame.Surface(self.screen.get_size()).convert()
        self.graph_surface = pygame.Surface(self.screen.get_size()).convert()

        self.grid_background_surface.set_colorkey(Color.BLACK.value)
        self.graph_surface.set_colorkey(Color.BLACK.value)

    def _update_background_surfaces(self, old_pos: Vector2, new_pos: Vector2):
        self.init_static_surfaces()
        self.graph_surface.fill(color=Color.BLACK.value)

        self.graph_drawer.draw_to_surface(self.graph_surface, self.grid_graph.graph, coordinate_converter=self.graph_to_coord, node_color=Color.ORANGE, enabled_edge_color=Color.BRIGHT_BLUE, disabled_edge_color=Color.GREY, draw_scale_matrix=self.get_grid_rotation_matrix())
        self.grid_drawer.draw_base_to_surface(self.grid_background_surface, self.grid, grid_color=Color.DARK_GREY, grid_draw_type=GridDrawType.LINES, draw_scale_matrix=self.get_grid_rotation_matrix())

    def _update_foreground_surfaces(self, old_pos: Vector2, new_pos: Vector2):
        self.init_dynamic_surfaces()
        self.grid_foreground_surface.fill(color=Color.BLACK.value)
        highlights = self.define_highlighted_grids(hover=new_pos)
        self.grid_drawer.draw_overlay_to_surface(self.grid_foreground_surface, self.grid, additional_highlight_grid_cells=highlights, draw_scale_matrix=self.get_grid_rotation_matrix())

        # self.unit_path_drawer.draw(self.grid, self.grid_drawer, self.grid_foreground_surface)

    def draw(self, frames:int, debug_mode: bool = False):
        #draw grid background
        if self.show_grid:
            self.screen.blit(self.grid_background_surface, (0, 0))

        #draw grid foreground
        if self.show_grid_highlights:
            self.screen.blit(self.grid_foreground_surface, (0, 0))

        #draw graph
        if self.show_graph:
            self.screen.blit(self.graph_surface, (0, 0))

        for entity in self.sprite_group:
            entity.blit(self.screen,
                        display_handle=lambda: self.debug_mode_toggle.value,
                        display_rect=lambda: self.debug_mode_toggle.value
                        )
            # self.screen.blit(entity.surf, entity.rect)
            


        if debug_mode:
            self.draw_mouse_coord(self.screen)
            self.draw_fps(self.screen, offset_rect=cRect(0, self.screen.get_height() - 100, 20, 100))
            self.draw_text(f"Sprite Count: {len(self.sprite_group)}", self.screen, offset_rect=Rectangle(0, 90, 100, 200))

        # Draw ASTAR Lines
        [self.curve_drawer.draw_curves({line: Color.YELLOW}, self.screen, draw_scale_matrix=self.get_grid_rotation_matrix()) for line in self._astar_curves or []]

        '''Draw Main Character'''
        self.main_character.blit(self.screen,
                                 display_handle=lambda: self.debug_mode_toggle.value,
                                 display_rect=lambda: self.debug_mode_toggle.value)


    def handle_hover_over(self, pressed_keys, mouse_pos_as_vector: Vector2):
        game_area_rect = cRect(0, 0, self.screen.get_height(), self.screen.get_width())
        self.grid_drawer.handle_hover_over(grid=self.grid, area_rect=game_area_rect,
                                           on_hover_changed_handlers=[self._update_foreground_surfaces, self.calculate_astar_from_cached_pos],
                                           draw_scale_matrix=self.get_grid_rotation_matrix())

    def handle_left_mouse_down(self, pressed_keys, mouse_pos):
        if pressed_keys[pygame.K_LCTRL]:
            grid_pos = self.grid_drawer.get_mouse_grid_pos(self.game_area_rectangle, help.mouse_pos_as_vector(), draw_scale_matrix=self.get_grid_rotation_matrix(), coord_to_grid_converter=self.grid.grid_from_coord)
            self.toggle_at_rc(grid_pos=grid_pos, disabler="BL")
            self._update_background_surfaces(Vector2(0, 0), Vector2(0, 0))
        else:
            name = f"b_{self.get_next_index()}"
            pos = help.mouse_pos_as_vector()
            self.agents[name] = pos
            self.sprite_group.add(RectangleSprite(name, pos, Color.BLUE, 30, 20))

    def handle_right_mouse_down(self, pressed_keys, mouse_pos):
        name = f"r_{self.get_next_index()}"
        pos = help.mouse_pos_as_vector()
        self.agents[name] = pos
        self.sprite_group.add(RectangleSprite(name, pos, Color.RED, 20, 30))

    def create_yellow(self):
        name = f"y_{self.get_next_index()}"
        pos = help.mouse_pos_as_vector()
        self.agents[name] = pos
        self.sprite_group.add(RectangleSprite(name, pos, Color.YELLOW, 20, 30))

    def model_updater(self, delta_time_ms: int):
        self.time_since_report_stats += delta_time_ms
        if self.time_since_report_stats > 10000:
            logging.info(self.tracked_time)
            self.time_since_report_stats = 0

        to_delete = []
        for agent, pos in self.agents.items():
            if agent[0] == "b":
                self.agents[agent] = pos + Vector2(0, 5)
            elif agent[0] == "r":
                self.agents[agent] = pos + Vector2(5, 0)
            elif agent[0] == "y":
                self.agents[agent] = pos + Vector2(5, 5)

            if pos.x > self.screen_width or pos.y > self.screen_height:
                to_delete.append(agent)

        for agent in to_delete:
            del self.agents[agent]


    def sprite_updater(self, delta_time_ms: int):
        char_pos = self.grid_drawer.grid_point_to_viewport_point(self.screen, self.grid, Vector2(0, 0), draw_scale_matrix=self.get_grid_rotation_matrix())
        char_pos = Vector2(char_pos.x, char_pos.y -256//4//2)
        self.main_character.set_pos(char_pos)

        '''animate'''
        self.animate_timer += delta_time_ms
        if self.animate_timer > self.animate_cycle:
            self.animate()

        for sprite in self.sprite_group:
            agent_pos = self.agents.get(sprite.id, None)
            if agent_pos is None:
                self.sprite_group.remove(sprite)
            else:
                sprite.set_pos(agent_pos)


    def animate(self):
        loop = True
        if self.main_character.current_animation == "fall":
            loop = False

        self.main_character.increment_animation_phase(loop)
        self.animate_timer = 0



    def calculate_astar_from_cached_pos(self, old_pos: Vector2, new_pos: Vector2):
        self._astar_curves = self.astar_curves(self.game_area_rectangle, gridGraph=self.grid_graph, start_pos=self._astar_start_point, end_pos=new_pos)

    def astar_curves(self, surface_rect: Rectangle, gridGraph: GridGraph, start_pos: Vector2, end_pos: Vector2):
        astr = [x.focal_point for x in gridGraph.astar_between_grid_pos(start_pos, end_pos).path]

        if astr is not None:
            astr_curves = []
            last = None
            for ii in range(0, len(astr)):
                if last is not None:
                    astr_curves.append(LineCurve(str(self.get_next_index()), gridGraph.grid.coord_from_grid_pos(last, surface_rect), gridGraph.grid.coord_from_grid_pos(astr[ii], surface_rect)))
                last = astr[ii]

            return astr_curves
        else:
            return None

    def set_astar_origin(self):
        self.set_astar_start_point(self.grid_drawer.hover_grid_pos)
        self._astar_curves = None

    def set_astar_start_point(self, grid_coords: Vector2):
        self._astar_start_point = grid_coords

    def define_highlighted_grids(self, hover:Vector2):
        highlights = {}
        grid_point = hover

        if grid_point.x < self.grid.nColumns and grid_point.x >= 0\
                and grid_point.y < self.grid.nRows and grid_point.y >= 0:
            highlights[grid_point] = Color.PURPLE

        # for col in range(0, self.grid.nColumns):
        #     for row in range(0,  self.grid.nRows):
        #         if self.grid.grid[row][col].state[self.toggle_name].value:
        #             highlights[Vector2(col, row)] = Color.GREEN

        return highlights

    def graph_to_coord(self, pos: Vector2):
        rectangle = self.game_area_rectangle
        return self.grid.coord_from_grid_pos(grid_pos=pos, area_rect=rectangle)

    def toggle_show_graph(self):
        self.show_graph = not self.show_graph

    def toggle_show_grid(self):
        self.show_grid = not self.show_grid

    def toggle_show_grid_highlights(self):
        self.show_grid_highlights = not self.show_grid_highlights

    def on_resize(self):
        self._update_background_surfaces(Vector2(0, 0), Vector2(0, 0))
        self._update_foreground_surfaces(Vector2(0, 0), Vector2(0, 0))

    def toggle_at_rc(self, grid_pos, disabler):
        if grid_pos.x < self.grid.nColumns and grid_pos.y < self.grid.nRows:
            self.grid.act_on_grid(grid_pos.y, grid_pos.x, [self.toggle_bool_policy])
            if self.grid.grid[grid_pos.x][grid_pos.y].state[self.toggle_name]:
                self.grid_graph.graph.enable_edges_to_node(Vector2(grid_pos.x, grid_pos.y), disabler)
            else:
                self.grid_graph.graph.disable_edges_to_node(Vector2(grid_pos.x, grid_pos.y), disabler)


            # self.grid.grid[grid_pos.y][grid_pos.x] = not self.grid.grid[grid_pos.y][grid_pos.x]
            # if self.grid.grid[grid_pos.y][grid_pos.x]:
            #     self.grid.enable_edges_to_node(Vector2(grid_pos.x, grid_pos.y), disabler)
            # else:
            #     self.grid.disable_edges_to_node(Vector2(grid_pos.x, grid_pos.y), disabler)



if __name__ == "__main__":
    import logging
    import loggingConfig

    loggingConfig.initLogging(loggingLvl=logging.INFO)
    game = MyGame()
    game.main()





from coopgame.characterDevelopment import character as char
import coopgame.characterDevelopment.stats as stat
from typing import Iterable
import logging

logger = logging.getLogger(__name__)


def acquire_traits(character: char.Character, traits: Iterable[stat.StatEffectPackage]):
    character.stats.register_events(events=[stat.StatsPackageStateEvent(description="Add Traits", added_stat_modifiers=traits)])
    logger.info(f"Character: {character} gained traits: {traits}")
    return character

def lose_traits(character: char.Character, traits: Iterable[stat.StatEffectPackage]):
    character.stats.register_events(events=[stat.StatsPackageStateEvent(description="Remove Traits", removed_stat_modifiers=traits)])
    logger.info(f"Character: {character} lost traits: {traits}")
    return character



import logging
from examples.voroni_world.models import *
from coopgame.gameTemplate import tt
from cooptools.decor import timer
import pprint
import math
from typing import Dict, Tuple, List, Callable
from cooptools import graphs as grph
from cooptools.physics import waypointFollower as wf
from examples.voroni_world.behaviors import cmn

logger = logging.getLogger(__name__)

_village_to_region_route_cache = {}

@timer(logger=logger, log_level=0, time_tracking_class=tt)
def choose_scoutable_region(village_region: Region,
                            regions: Dict[int, Region],
                            route_provider: Callable[[Region, Region], grph.AStarResults]) -> Tuple[List[Region], Region]:
    scoutables = [r for r in regions.values()
                  if r.visibility_level == VisibilityLevel.REVEALABLE]

    routes = {}
    for r in scoutables:
        if r in _village_to_region_route_cache:
            routes[r] = _village_to_region_route_cache[r]
        else:
            routes[r] = route_provider(village_region, r)

            if routes[r].nodes is not None:
                _village_to_region_route_cache[r] = routes[r]

    route_ls = {k.name: v.Length if v.nodes is not None else None
                for k, v in routes.items() if v.nodes is not None}
    logger.debug(f"Scoutable Regions: {pprint.pformat(route_ls)}")

    """CRASH IN"""
    r_654 = next((x for x in scoutables if x.name == 654), None)
    if r_654 is not None: print(routes[r_654].nodes)
    # if r_654 is not None and routes[r_654].nodes is None:
    #     osm.save_jsonable_data(routes[r_654].to_json(), file_path=r"C:\Users\Tj Burns\Downloads\654_route.json")
    #     grph.plot_astar_steps(
    #         results=routes[r_654],
    #         graph=self.region_graph,
    #         save_path=r"C:\Users\Tj Burns\Downloads\654_route.gif"
    #     )

    if len(route_ls) == 0:
        return scoutables, None

    """CRASH OUT"""

    min_l = min(l for l in route_ls.values())
    selct = next((k, v) for k, v in routes.items()
                 if v.nodes is not None and math.isclose(v.Length, min_l))

    del _village_to_region_route_cache[selct[0]]

    logger.debug(f"Selected Region to scout: {selct[0]} with dist {selct[1].Length}")
    return scoutables, selct[0]

def reveal_a_region(r: Region,
                    regions: Dict[int, Region],
                    force: bool = False):

    r.set_visibility(requested=VisibilityLevel.REVEALED, force=force)
    r.remove_action_state(VisibilityLevel.REVEALING, force=force)

    for neighbor in r.neighbors.keys():
        if regions[neighbor].visibility_level in [VisibilityLevel.HIDDEN]:
            regions[neighbor].set_visibility(VisibilityLevel.REVEALABLE)

    return r

def scout(
        villager: Villager,
        villager_current_region_provider: Callable[[Villager], Region],
        route_provider: Callable[[Region, Region], grph.AStarResults],
        route_to_waypoints_provider: Callable[[List[grph.Node]], List[wf.Waypoint]],
        village_region: Region,
        regions: Dict[int, Region],
):

    scoutables, to_scout = choose_scoutable_region(
        village_region=village_region,
        regions=regions,
        route_provider=route_provider
    )

    if to_scout is None:
        logger.error(f"No routable, scoutable areas")
        return None

    if to_scout.visibility_level != VisibilityLevel.REVEALABLE:
        raise Exception(
            f"Villager \'{villager.name}\' cannot scout region \'{to_scout.name}\' because its visibility is {to_scout.visibility_level}")

    cmn.send_villager_to_region(
        villager=villager,
        region=to_scout,
        villager_current_region_provider=villager_current_region_provider,
        route_provider=route_provider,
        route_to_waypoints_provider=route_to_waypoints_provider,
        stop_at_border=True
    )

    to_scout.set_visibility(requested=VisibilityLevel.REVEALING)
    to_scout.add_action_state(VisibilityLevel.REVEALING)

    return to_scout


import time

from cooptools.register import Register
from coopgame.characterDevelopment import character as char
from typing import Iterable
from cooptools.typeProviders import StringProvider, resolve_string_provider
from cooptools.common import verify_val
import coopgame.characterDevelopment.stats as stat

class Community:
    def __init__(self,
                 name: str,
                 character_generator_options: Iterable[char.CharacterGeneratorArgs],
                 init_pop_size: int = None,
                 init_pop: Iterable[char.Character] = None):
        self.name: str = name
        self.character_registry = Register[char.Character]()
        self.global_effects_registry = Register[stat.StatEffectPackage]()
        self.char_generator_options = character_generator_options

        if init_pop_size is not None: self.add_new_characters(count=init_pop_size)
        if init_pop is not None: self.add_new_characters(new_characters=init_pop)

    def add_new_characters(self,
                           count: int = None,
                           new_characters: Iterable[char.Character] = None):
        if new_characters is not None:
            self.character_registry.register(new_characters, ids=[x.id for x in new_characters])

        if count is not None:
            verify_val(count, gte=0, error_msg=f"New character count {count} is invalid")
            self.add_new_characters(new_characters=[
                char.Character().generate()
                    args=rnd.choice(list(self.char_generator_options)))
                for x in range(count)
            ])




    def __repr__(self):
        chars_txt = "\n\t" + "\n\t".join([f"{ii}: {x}" for ii, x in enumerate(self.character_registry.Registry.values())])
        return f"Community Name: {self.name}" \
               f"\nCharacters: {chars_txt}"


if __name__ == "__main__":
    from cooptools.randoms import a_string
    import random as rnd
    import coopgame.characterDevelopment.stat_modifier_manifest as smm
    from coopgame.characterDevelopment.behaviorExtensions import traits as traits

    character_naming_provider = lambda: a_string(10)

    default_stats = [
        stat.StatState('hp', val=1000, meta=stat.StatMeta(0, 1000)),
        stat.StatState('mana', val=100, meta=stat.StatMeta(0, 100)),
        stat.StatState('rage', val=0, meta=stat.StatMeta(0, 100))
    ]

    config_fn_provider = r'C:\Users\tburns\PycharmProjects\coopgame\coopgame\characterDevelopment\manifest.json'
    char_gen_options = [
        char.CharacterGeneratorArgs(
            name_provider=character_naming_provider,
            init_stats_package=stat.StatsPackageState(
                base_stats=stat.StatStates({x.stat_name: x for x in default_stats})
            ),
            traits_manifest=stat.StatEffectPackageManifest.from_config_data(config_fn_provider)
        ) for k, v in smm.races_manifest.items()
    ]

    my_community = Community(name='c1',
                             character_generator_options=char_gen_options,
                             init_pop_size=10)

    active_global_events = []

    # Eventually, this should become some sort of "World" class
    epoch = 0
    while True:
        epoch += 1
        p_global_event_occurs = 0.35
        p_global_event_ends = 0.15
        p_char_gain_personality_trait = 0.10
        p_char_gain_physical_trait = 0.05
        p_char_lose_personality_trait = 0.01
        p_char_lose_physical_trait = 0.01

        # handle global event(s)
        for x in active_global_events:
            if rnd.random() < p_global_event_ends:
                active_global_events.remove(x)

        if rnd.random() < p_global_event_occurs:
            active_global_events.append(smm.global_effect_manifest[rnd.choice(list(smm.GlobalEffect))])

        for k, v in my_community.character_registry.Registry.items():
            # add new personality traits:
            if rnd.random() < p_char_gain_personality_trait:
                traits.acquire_traits(v,
                                      traits=[smm.personality_trait_manifest[rnd.choice(list(smm.PersonalityTrait))]])

            # add new physical traits:
            if rnd.random() < p_char_gain_personality_trait:
                traits.acquire_traits(v,
                                      traits=[smm.physical_trait_manifest[rnd.choice(list(smm.PhysicalTrait))]])

        print(my_community)
        time.sleep(1)





from coopgame.gameTemplate import GameTemplate
import coopgame.pygamehelpers as help
from cooptools.colors import Color
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from cooptools.transform import Transform
from typing import Tuple

SCROLL_SPEED = 5
ZOOM_SCALE = 0.75

class TransformGame_Basic(GameTemplate):

    def __init__(self):
        self._camera = Transform(scale=[100, 100, 100])
        super().__init__(input_state_handler=self.define_input_handler())

    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_w, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(0, SCROLL_SPEED, 0))),
                InputAction([InputEvent(PyKeys.K_a, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(-SCROLL_SPEED, 0, 0))),
                InputAction([InputEvent(PyKeys.K_s, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(0, -SCROLL_SPEED, 0))),
                InputAction([InputEvent(PyKeys.K_d, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(SCROLL_SPEED, 0, 0))),
                InputAction([InputEvent(PyKeys.K_UP, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(0, SCROLL_SPEED, 0))),
                InputAction([InputEvent(PyKeys.K_LEFT, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(-SCROLL_SPEED, 0, 0))),
                InputAction([InputEvent(PyKeys.K_DOWN, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(0, -SCROLL_SPEED, 0))),
                InputAction([InputEvent(PyKeys.K_RIGHT, event_type=InputEventType.HELD)], callback=lambda x: self._camera.Translation.update(delta_vector=(SCROLL_SPEED, 0, 0))),
                InputAction([InputEvent(PyMouse.SCROLLUP, event_type=InputEventType.DOWN)], callback=lambda x: self._camera.Scale.update(scalar=1/ZOOM_SCALE)),
                InputAction([InputEvent(PyMouse.SCROLLDOWN, event_type=InputEventType.DOWN)], callback=lambda x: self._camera.Scale.update(scalar=ZOOM_SCALE)),
        ])

    def draw(self, frames: int, debug_mode: bool = False):
        help.draw_polygon(self.screen,
                          points=[
                              (0, 1, 0),
                              (1, 0, 0),
                              (0.75, -1, 0),
                              (-0.75, -1, 0),
                              (-1, 0, 0)
                          ],
                          color=Color.YELLOW,
                          draw_transform_matrix=self._camera.transform_matrix(
                              inversions=['Y']
                          )
                          )

if __name__ == "__main__":
    g = TransformGame_Basic()
    g.main()


from coopgame.harfang.harfangGameTemplate import HarfangGameTemplate
from coopgame.harfang.harfangPrimitives import draw_circle, draw_triangle, draw_square
from coopgame.harfang.harfangColorConverter import harfang_color_from_coopcolor
from cooptools.colors import Color
from coopstructs.vectors import Vector2
from noise import snoise2, pnoise3
from coopbugger.monitoredclass import MonitoredClass


REZ = 15
MAX_SIZE = 20
PARTICLE_COLOR = Color.DARK_BLUE

class Particle():
    def __init__(self, pos: Vector2, value, size, color = None):
        self.pos = pos
        self.size = size
        self.thickness = 1
        self.value = value
        self.color = color

    def display(self, color: Color = None, **kwargs):
        if color is None:
            color = self.color
        # draw_triangle(kwargs.get('plus'),
        #               self.pos + Vector2(-self.size / 2, -self.size / 2),
        #               self.pos + Vector2(self.size / 2, -self.size / 2),
        #               self.pos + Vector2(0, self.size / 2),
        #               harfang_color_from_coopcolor(coopcolor=color))
        draw_square(kwargs.get('plus'),
                    self.pos,
                    self.size)

class HarfangPNoise(HarfangGameTemplate, MonitoredClass):
    def initialize_game(self):
        pass

    def handle_right_click(self, pressed_keys):
        pass

    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        pass

    def handle_mouse_scroll_up(self, pressed_keys):
        pass

    def handle_mouse_scroll_down(self, pressed_keys):
        pass

    def model_updater(self, delta_time_ms: int):
        pass

    def sprite_updater(self, delta_time_ms: int):
        pass

    def on_resize(self):
        pass

    def register_button(self, id, text, callback, postion_rect):
        pass

    def register_keys(self):
        pass

    def __init__(self):
        HarfangGameTemplate.__init__(self)
        MonitoredClass.__init__(self)
        self.cols = int(self.screen_width / REZ + 1)
        self.rows = int(self.screen_height /REZ + 1)

        self.xoff = 0
        self.yoff = 0
        self.zoff = 0
        self.xyincrement = .002
        self.zincrement = 0

        self.noise_scale = max(self.rows, self.cols)/5
        self.particles = []

        self._blitables = {}
        self.show_overlap = False


        self.create_particles()

        self.line_surface_layers = []
        self.lines_surface = None
        self.lines_surface_offset = Vector2(0, 0)

    @MonitoredClass.timer
    def draw(self, frames:int, debug_mode: bool = False):
        self.xoff += self.xyincrement
        self.yoff += self.xyincrement
        self.zoff += self.zincrement

        self.update_particles()

        for row in self.particles:
            for particle, color in row:
                particle.display(plus=self.plus, color=color)

        self._draw_fps(self.runtime_config)

    def handle_left_click(self, pressed_keys):
        self.show_overlap = not self.show_overlap

    @MonitoredClass.timer
    def update_particles(self):

        if self.particles is None or len(self.particles) == 0:
            self.create_particles()

        octaves = 6
        persistence = .5
        lacunarity = 2


        for ii in range(0, self.cols):
            # self.particles.append([])
            for jj in range(0, self.rows):
                # color = (rnd.randint(0, 255), rnd.randint(0, 255), rnd.randint(0, 255))
                value = pnoise3(
                    (ii) / self.noise_scale + self.xoff,
                    (jj) / self.noise_scale + self.yoff,
                    self.zoff,
                    octaves=octaves,
                    persistence=persistence,
                    lacunarity=lacunarity,
                    repeatx=1024,
                    repeaty=1024,
                    base=0)
                # color = Color.closest_color((0, 0, int(255 * (value + 1 / 2))))
                color = (0, 0, int(255 * (value + 1 / 2)))

                self.particles[ii][jj][0].value = value
                self.particles[ii][jj][0].size = max(int((value + 1 / 2) * MAX_SIZE), 1)
                self.particles[ii][jj][0].color = color



    @MonitoredClass.timer
    def create_particles(self):
        self.particles.clear()

        for ii in range(0, self.cols):
            self.particles.append([])
            for jj in range(0, self.rows):
                self.particles[ii].append(
                    (Particle(Vector2(ii * REZ, jj * REZ), 0, 1), PARTICLE_COLOR))

    # @MonitoredClass.timer
    # def draw_lines_to_surface(self, surface: pygame.Surface, only_last=False):
    #     counter = 0
    #     for ii in range(0, self.cols-1):
    #         for jj in range(0, self.rows-1):
    #             if (only_last and not((ii == self.cols - 3) or (jj == self.rows - 3))):
    #                 continue
    #
    #             vals = self.particles[ii][jj][0].value, self.particles[ii+1][jj][0].value, self.particles[ii+1][jj+1][0].value, self.particles[ii][jj+1][0].value
    #             state = self.get_state_from_square(*[math.ceil(x) for x in vals])
    #             lines = self.line_state_switch(state)
    #             crit_points = self.crit_points_from_square(self.particles[ii][jj][0].pos, REZ, *vals)
    #             # crit_points = self.square_crits_of_point(self.particles[ii][jj][0].pos, REZ)
    #             for line in lines:
    #                 pygame.draw.line(surface, Color.WHITE.value, (crit_points[line[0]].x, crit_points[line[0]].y), (crit_points[line[1]].x, crit_points[line[1]].y))
    #                 counter += 1
    #
    #     logging.debug(f"{counter} lines drawn")
    #
    # @MonitoredClass.timer
    # def draw_lines(self):
    #     if self.line_surface_layers is not None and len(self.line_surface_layers) > 0:
    #         print(len(self.line_surface_layers))
    #         for surface, offset in self.line_surface_layers:
    #             self.screen.blit(surface, (offset.x, offset.y))
    #     else:
    #         self.draw_lines_to_surface(self.screen)


    @MonitoredClass.timer
    def get_state_from_square(self, a, b, c, d):
        # return 1*a + 2*b +4*c+8*d
        return 8 * a + 4 * b + 2 * c + 1 * d


    @MonitoredClass.timer
    def line_state_switch(self, state):
        switcher = {
            0:[],
            1:[(2, 3)],
            2:[(1, 2)],
            3:[(1, 3)],
            4:[(0, 1)],
            5:[(0, 3), (1, 2)],
            6:[(0, 2)],
            7:[(0, 3)],
            8:[(0, 3)],
            9:[(0, 2)],
            10:[(0, 1), (2, 3)],
            11:[(0, 1)],
            12:[(1, 3)],
            13:[(1, 2)],
            14:[(2, 3)],
            15:[],
        }
        return switcher.get(state, [])

    @MonitoredClass.timer
    def crit_points_from_square(self, point: Vector2, resolution, val1, val2, val3, val4):
        """:returns a tuple of the critical points based on the position of the top left of the square and the
        values of the 4 corners of the square. ordered TL, TR, BR, BL (clockwise starting from top left)
        values assumed to be between -.5 and .5"""

        val1 +=1
        val2 +=1
        val3 +=1
        val4 +=1


        amtA = (1 - val1)/(val2 - val1) if val1 != val2 else .5
        a = Vector2(point.x + resolution * amtA, point.y)
        # a = point.interpolate(point+Vector2(resolution, 0), amtA)
        amtB = (1 - val2)/(val3 - val2) if val3 != val2 else .5
        b = Vector2(point.x + resolution, point.y + resolution * amtB)
        # b = (point+Vector2(resolution, 0)).interpolate(point + Vector2(resolution, resolution), amtB)
        amtC = (1 - val3) / (val4 - val3) if val3 != val4 else .5
        c = Vector2(point.x + resolution * (1 - amtC),  point.y + resolution)
        # c = (point + Vector2(resolution, resolution)).interpolate(point + Vector2(0, resolution), amtC)
        amtD = (1 - val4) / (val1 - val4) if val1 != val4 else .5
        d = Vector2(point.x, point.y + resolution * (1 - amtD))
        # d = (point + Vector2(0, resolution)).interpolate(point, amtD)

        return (a, b, c, d)


    @MonitoredClass.timer
    def square_crits_of_point(self, point:Vector2, resolution):
        a = Vector2(point.x + resolution/2  , point.y)
        b = Vector2(point.x + resolution    , point.y + resolution/2)
        c = Vector2(point.x + resolution/2  , point.y + resolution)
        d = Vector2(point.x                 , point.y + resolution/2)
        return (a, b, c, d)


if __name__ == "__main__":
    import logging
    logging.basicConfig(level=logging.INFO)
    game = HarfangPNoise()
    game.main()


from coopgame.characterDevelopment import character as char
import coopgame.characterDevelopment.stats as stat

MANA = 'MANA'
MAX_MANA = 'MAX_MANA'

class CastSpellCharacterEvent(stat.DeltaStatStatsPackageStateEvent):
    def __init__(self,
                 spell_name: str,
                 spell_cost: float,
                 mana_stat_name: str = MANA):
        super().__init__(
            description=f"Cast Spell \"{spell_name}\" for {spell_cost} mana",
            stat_name=mana_stat_name,
            delta=-spell_cost
        )

def current_mana(character: char.Character,
               mana_stat_name: str = MANA) -> float:
    return character.stats.State.resolve_stat(stat_name=mana_stat_name).val

def max_mana(character: char.Character,
           max_mana_stat_name: str = MAX_MANA) ->float:
    return character.stats.State.resolve_stat(stat_name=max_mana_stat_name).val

def missing_mana(character: char.Character,
               mana_stat_name: str = MANA,
               max_mana_stat_name: str = MAX_MANA):
    current = current_mana(character, mana_stat_name=mana_stat_name)
    max = max_mana(character, max_mana_stat_name=max_mana_stat_name)

    return max - current

def cast_spell(character: char.Character,
               spell_name: str,
               mana_cost: float,
               mana_stat_name: str = MANA):
    current = current_mana(character=character,
                           mana_stat_name=mana_stat_name)

    if mana_cost > current:
        raise ValueError(f"Tried to cast spell {spell_name}, but not enough mana. Current: {current}, Cost: {mana_cost}")

    character.stats.register_events(events=[CastSpellCharacterEvent(spell_name=spell_name, spell_cost=mana_cost)])
    return character


def restore(character: char.Character,
         amount: float = None,
         mana_stat_name: str = MANA,
         max_mana_stat_name: str = MAX_MANA
         ):
    if amount is None:
        amount = missing_mana(character=character,
                            mana_stat_name=mana_stat_name,
                            max_mana_stat_name=max_mana_stat_name)

    character.stats.register_events(events=[stat.DeltaStatStatsPackageStateEvent(stat_name=mana_stat_name, delta=amount)])
    return character




from typing import Tuple, Callable, Dict, Any
import math
import numpy as np
import pprint
from cooptools.sectors import sec_u as sec_util

coord_type_resolver= Callable[[Tuple[int, int]], Any]

def get_chunk(target: Tuple[int, int],
               chunk_size: Tuple[int, int],
               tile_type_resolvers: Dict[str, coord_type_resolver]) -> Dict[Tuple[int, int], Dict[str, Any]]:
    chunk_data = {}
    for row in range(chunk_size[1]):
        for col in range(chunk_size[0]):
            target_col = target[1] * chunk_size[1] + col
            target_row = target[0] * chunk_size[0] + row
            coord = (target_col, target_row)

            state = {}
            for type, resolver in tile_type_resolvers.items():
                state[type] = resolver(coord)

            chunk_data[(target_row, target_col)] = state

    return chunk_data

def chunk_of_tile(tile_coord, chunk_shape):
    return (int(math.floor(tile_coord[0] / chunk_shape[0])), int(math.floor(tile_coord[1]/ chunk_shape[1])))


class ChunkRenderer:

    def __init__(self,
                 chunk_size: Tuple[int, int],
                 tile_type_resolvers: Dict[str, coord_type_resolver]):

        """The chunk data is a dictionary that contains k: the chunk coord, v: the list of tile types at coord within
        the chunk"""
        self.chunk_data: Dict[Tuple[int, int], Dict[Tuple[int, int], Dict[str, Any]]] = {}
        self.chunk_size: Tuple[int, int] = chunk_size
        self.tile_type_resolvers: Dict[str, coord_type_resolver] = tile_type_resolvers

    def generate_get_chunk_at_coord(self, coord):
        if coord not in self.chunk_data:
            self.chunk_data[coord] = get_chunk(coord, self.chunk_size, self.tile_type_resolvers)
        return self.chunk_data[coord]

    def tile_pxls(self, chunk_pxls):
        return sec_util.sector_dims(area_dims=chunk_pxls, sector_def=self.chunk_size)

    def renderable_tile_type_array(self,
                                   scroll: Tuple[float, float],
                                   window_size: Tuple[int, int],
                                   chunk_pxls: Tuple[int, int],
                                   key: str,
                                   buffer_tiles: Tuple[int, int] = None
                                   ) -> np.ndarray:
        # get the size of the tile array that can be rendered.
        tile_pxls = self.tile_pxls(chunk_pxls)
        visible_tile_shape = sec_util.sectors_in_window(window_dims=window_size,
                                          sector_dims=tile_pxls,
                                          buffer_sectors=(buffer_tiles[0], buffer_tiles[1], 0, 0))

        # init a zero array of the shape of the renderable tiles
        ret = np.zeros(shape=visible_tile_shape, dtype=int)

        rows, cols = sec_util.sectors_in_window(window_dims=window_size,
                                                sector_dims=tile_pxls,
                                                buffer_sectors=(buffer_tiles[0], buffer_tiles[1], 0, 0))
        # get origin tile
        o_tile = sec_util.sector_from_coord(
            coord=scroll,
            # sec_dims=tile_pxls
            sector_def=(rows, cols),
            area_dims=window_size
        )

        # fill the array with the values of
        for tile_rc in self.visible_tile_generator(scroll, window_size, chunk_pxls, buffer_tiles):
            chunk_coord = chunk_of_tile(tile_rc, self.chunk_size)
            chunk_data = self.generate_get_chunk_at_coord(chunk_coord)
            type = chunk_data[tile_rc]
            ret_coord = (tile_rc[0] - o_tile[0], tile_rc[1] - o_tile[1])
            ret[ret_coord] = type[key]

        return ret

    def visible_tile_idx_by_condition(self,
                                      scroll: Tuple[float, float],
                                      window_size: Tuple[int, int],
                                      chunk_pxls: Tuple[int, int],
                                      tile_evaluator: Callable[[Dict], bool],
                                      buffer_tiles: Tuple[int, int] = None) -> Dict[Tuple[int, int], Dict]:

        ret = {}
        # fill the array with the values of
        for tile_rc in self.visible_tile_generator(scroll, window_size, chunk_pxls, buffer_tiles):
            state = self.tile_state(tile_rc)
            if tile_evaluator(state):
                ret[tile_rc] = state

        return ret

    def visible_chunk_generator(self,
                                scroll: Tuple[float, float],
                                window_size: Tuple[int, int],
                                chunk_pxls: Tuple[int, int]):
        renderable_chunk_shape = sec_util.sectors_in_window(window_size, chunk_pxls, scroll) #self.renderable_chunk_shape(window_size, chunk_pxls, scroll)
        o_chunk_row, o_chunk_col = sec_util.sector_from_coord(scroll, chunk_pxls)

        for col in range(renderable_chunk_shape[1]):
            for row in range(renderable_chunk_shape[0]):
                chunk_row = row + o_chunk_row
                chunk_col = col + o_chunk_col
                yield (chunk_row, chunk_col)

    def visible_tile_generator(self,
                               scroll: Tuple[float, float],
                               window_size: Tuple[int, int],
                               chunk_pxls: Tuple[int, int],
                               buffer_tiles: Tuple[int, int] = None
                               ):

        tile_pxls = sec_util.sector_dims(chunk_pxls, self.chunk_size)
        rows, cols = sec_util.sectors_in_window(window_dims=window_size,
                                          sector_dims=tile_pxls,
                                          buffer_sectors=(buffer_tiles[0], buffer_tiles[1], 0, 0))
        o_tile_coord = sec_util.sector_from_coord(coord=scroll,
                                                  area_dims=window_size,
                                                  sector_def=(rows, cols))
        for col in range(cols):
            for row in range(rows):
                tile_row = row + o_tile_coord[0]
                tile_col = col + o_tile_coord[1]
                yield (tile_row, tile_col)

    def tile_state(self, tile: Tuple[int, int]) -> Dict[str, Any]:
        tile_chunk = chunk_of_tile(tile, self.chunk_size)
        return self.generate_get_chunk_at_coord(tile_chunk)[tile]


if __name__ == "__main__":
    import random as rnd

    chunk_size = (8, 8)
    window_size = (1500, 700)


    def dummy_tile_type_resolver(target: Tuple[int, int]) -> int:
        if target[1] > 10:
            return 2
        if target[1] == 10:
            return 1
        if target[1] == 9:
            return rnd.randint(1, 5) == 1

        return 0





from coopgame.gameTemplate import GameTemplate
import coopgame.pygamehelpers as help
from cooptools.colors import Color
from coopgame.grids_graphs.pygridhandler import PyGridHandler
from cooptools.toggles import BooleanToggleable
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from cooptools.sectors.grids import RectGrid
from cooptools.dictPolicies import TogglePolicy
from coopstructs.geometry.vectors.vectorN import Vector2
from coopgame.grids_graphs.pygridhandler import GridSurfaceType


ROWS = 100
COLS = 200
# GRID_LINE_COLOR = Color.CHARCOAL
# CENTER_GUIDE_COLOR = Color.DARK_GREY
# HOVER_GRIDBOX_COLOR = Color.YELLOW
# CROSSHAIRS_COLOR = Color.BURLY_WOOD
STATIC_HIGHLIGHTED_CELLS = {Vector2(1, 1): Color.PURPLE,
                            Vector2(5, 5): Color.YELLOW}
GRID_TOGGLE_POLICY = TogglePolicy(key='toggled', toggle_type=BooleanToggleable, default=False)

class GridGame_Basic(GameTemplate):

    def __init__(self):
        super().__init__(input_state_handler=self.define_input_handler())

        self.grid = RectGrid(nRows=ROWS, nColumns=COLS)
        self.grid_handler = PyGridHandler(
            grid_provider=lambda: self.grid,
            screen=self.screen,
            get_game_area_rect_callback=lambda: self.game_area_rectangle,
            get_highlights_callback=lambda: STATIC_HIGHLIGHTED_CELLS,
            get_highlighted_rows=lambda: self.highlighted_rows(),
            get_highlighted_cols=lambda: self.highlighted_cols(),
            get_toggled_key_colors=lambda: {GRID_TOGGLE_POLICY.key: Color.RED}
        )

    def highlighted_rows(self):
        return {
            3: Color.ANTIQUE_WHITE
        }
    def highlighted_cols(self):
        return {
            3: Color.PEACH_PUFF
        }

    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_r, event_type=InputEventType.DOWN)], callback=lambda x: self.redraw()),
                InputAction([InputEvent(PyKeys.K_1, event_type=InputEventType.DOWN)], callback=lambda x: self.grid_handler.toggle_surface([GridSurfaceType.GRID_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_2, event_type=InputEventType.DOWN)], callback=lambda x: self.grid_handler.toggle_surface([GridSurfaceType.HIGHLIGHTS_SURFACE_ID])),
                InputAction([InputEvent(PyMouse.LEFT, event_type=InputEventType.DOWN)], callback=lambda x: self.toggle_grid_at_mouse_pos()),
                InputAction([InputEvent(PyKeys.K_KP_PLUS, event_type=InputEventType.DOWN)], callback=lambda x: self.grid_handler.show_all()),
                InputAction([InputEvent(PyKeys.K_KP_MINUS, event_type=InputEventType.DOWN)], callback=lambda x: self.grid_handler.hide_all()),
        ])

    def toggle_grid_at_mouse_pos(self):
        mgpo = self.grid_handler.mouse_grid_pos()
        self.grid.act_at_loc(row=int(mgpo.x), column=int(mgpo.y), policies=[GRID_TOGGLE_POLICY])
        self.grid_handler.invalidate()

    def redraw(self):
        self.grid_handler.redraw()

    def draw(self, frames: int, debug_mode: bool = False):
        self.grid_handler.update()
        self.grid_handler.render(self.screen)


if __name__ == "__main__":
    g = GridGame_Basic()
    g.main()


from dataclasses import dataclass, asdict
from typing import Dict, List, Iterable
import random as rnd
from cooptools.config import JsonConfigHandler
from cooptools.typeProviders import StringProvider, resolve_string_provider

@dataclass(frozen=True)
class MutuallyExclusiveManifest:
    options: Dict[str, List]

    def random(self,
               type: str = None,
               excluded_types: Iterable[str] = None
               ):

        available_types = [x for x in self.options.keys() if excluded_types is None or x not in excluded_types]
        if len(available_types) == 0:
            return None, None

        if type is None: type = rnd.choice(available_types)

        return (type, rnd.choice(self.options[type]))

    @classmethod
    def from_config(cls,
                    config_file_name_provider: StringProvider,
                    key: str):
        config = JsonConfigHandler(file_path_provider=config_file_name_provider)
        data = config.resolve(key)

        return MutuallyExclusiveManifest(
            options=data
        )





if __name__ == '__main__':
    manifest = MutuallyExclusiveManifest(
        options={
            'beauty': 'Ugly / Plain / Attractive / Beautiful'.split(' / '),
            'strength': 'Puny / Weak / Strong / Mighty'.split(' / '),
            'height': 'Dwarf / Short / Tall / Giant'.split(' / '),
            'weight': 'Skinny / Thin / Stout / Obese'.split(' / ')
        }
    )

    used_types = []
    for i in range(10):
        type, choice = manifest.random(excluded_types=used_types)
        used_types.append(type)
        print(type, choice)






import logging
import uuid

from cooptools.timeTracker import TimeTracker
# from cooptools import taskProcessing as tsk
from cooptools import graphs as grph
from cooptools.physics import waypointFollower as wf
from cooptools.sectors import SectorTree
from voroni import *
from wave_collapse import *
from typing import Iterable
import math
import pprint
from cooptools import os_manip as osm
from cooptools.coopEnum import CoopEnum
from cooptools.toggles import BooleanToggleable
from examples.voroni_world.models import *
from behaviors import *
from input_controller import InputController
from behaviorTypes import BehaviorType
import cooptools.typeProviders as tp
import coopstorage.storage2 as store
from cooptools import randoms as crnd

logger = logging.getLogger(__name__)


class ClientType(CoopEnum):
    VILLAGER = auto()
    REGION = auto()


class RoadType(CoopEnum):
    DIRT_PATH = .9
    COBBLESTONE_ROAD = .8
    PAVED_ROAD = .5
    RAIL = .25


@timer(logger=logger, log_level=0, time_tracking_class=tt)
def _generate_regions(n, bounds) -> Tuple[List[Region], Voronoi, Dict[int, List[int]]]:
    pts = [pt for pt in vec.random_point_generator(n, bounds)]

    # generate voronoi diagram
    vor, vn_map, points, polys, border_map = generate_voronoi(points=pts)

    terrains = wave_collapse_terrains(vn_map)

    regions = []
    neighbor_cnxn_node_ids: Dict[frozenset, str] = {}


    for ii, pt in enumerate(pts):
        logging.debug(f"Adding region {ii + 1} / {len(pts)}")

        neigh_cnxns = {}
        for neigh in vn_map[ii]:
            if (id_neigh := frozenset([ii, neigh])) not in neighbor_cnxn_node_ids:
                neighbor_cnxn_node_ids[id_neigh] = f"{ii}<>{neigh}"

            neigh_cnxns[neigh] = neighbor_cnxn_node_ids[id_neigh]

        regions.append(
            Region(
                name=ii,
                pt=pt,
                terrain=terrains[ii],
                boundary=polys[ii],
                vert_node_ids=vor.regions[ii],
                neighbors={
                    x: NeighborAttrs(
                        shared_verts=border_map[frozenset([ii, x])],
                        cnxn_node_id=f"C:{neigh_cnxns[x]}",
                    ) for x in vn_map[ii]
                }
            )
        )

    return regions, vor, vn_map


class GameModel:

    def __init__(self,
                 input_controller: InputController,
                 max_villagers: int,
                 init_villagers: int,
                 villager_spawn_ms: int,
                 sector_tree_depth: int = 3,
                 seconds_per_hr: float = 1):
        self.input_controller = input_controller

        self.regions: Dict[int, Region] = None
        self.world_sector_map: SectorTree = None
        self.village_id: int = None
        self.world_graph: grph.Graph = None
        self.world_clock: TimeTracker = TimeTracker(as_async=False, start_on_init=True).start()
        self._seconds_per_hr = seconds_per_hr
        self._storage = store.Storage.from_meta(
            location_type_counts=[(store.dcs.LocationMeta(
                dims=(10, 10, 10),
                channel_processor=store.FIFOFlowChannelProcessor(),
                capacity=3
            ), 100)],
            naming_provider=lambda x: crnd.a_string(k=7)
        )

        self._last_villager_name = -1

        self.villager_spawn_timer = TimeTracker(timeout_ms=villager_spawn_ms,
                                                start_on_init=True,
                                                as_async=False,
                                                id="Villager Spawn Timer")
        self._villagers: Dict[int, Villager] = {}

        self._max_villagers = max_villagers
        self._n_init_villagers = init_villagers

        self._villager_waypoint_cache = {}

        self.crash_toggle: BooleanToggleable = BooleanToggleable(default=False)
        self.crash_regions: Tuple[Region, Region] = None

        self.rivers: grph.Graph = None
        # self._tasks = List[tsk.Task]
        self._sector_tree_depth = sector_tree_depth

    def choose_village_region(self) -> Region:
        village = next(x for x in self.shuffled_region_generator() if x.terrain in [
            TerrainType.PLAINS,
            TerrainType.GRASS,
            TerrainType.PLAINSHILLS,
            TerrainType.GRASSHILLS,
            TerrainType.BEACH
        ])

        return reveal_a_region(village, regions=self.regions, force=True)

    def create_villager(self):
        name = self._get_villager_name()
        self._villagers[name] = Villager(
            name=name).set_pos(self.Village.pt)

        self.world_sector_map.add_update_clients({self._villagers[name].id: ClientType.VILLAGER})
        self.villager_spawn_timer.reset(start=True)

    def init_villagers(self, n):
        logger.info(f"Initializing {n} villagers")
        self._villagers = {}
        self._last_villager_name = -1
        for ii in range(n):
            self.create_villager()

    def _build_boundary_graph(
            self,
            regions: Dict[int, Region]
    ) -> grph.Graph:
        logger.debug(f"Building Boundary Graph...")
        graph_dict = {}

        for id, region in regions.items():
            me = grph.Node(region.node_id, region.pt)

            for neigh, attrs in region.neighbors.items():
                if attrs.Boundary is not None:
                    cnxn_pos = attrs.Boundary.Midpoint.as_tuple()
                else:
                    cnxn_pos = tuple(float(x) for x in list(attrs.shared_verts.values())[0])

                shared_vert_nodes = []
                for shared_vert, vert_pos in attrs.shared_verts.items():
                    shared_vert_node = grph.Node(
                        name=f"V{shared_vert}",
                        pos=tuple(x for x in vert_pos)
                    )
                    shared_vert_nodes.append(shared_vert_node)

                cnxn_pos_node = grph.Node(
                    name=region.neighbors[neigh].cnxn_node_id,
                    pos=cnxn_pos
                )


                # Connect first shared vert to the in-between cnxn node
                graph_dict.setdefault(shared_vert_nodes[0], []).append(cnxn_pos_node)

                # Connect region center to the cnxn node
                graph_dict.setdefault(me, []).append(cnxn_pos_node)

                # connect cnxn node back to region and the first shared node
                graph_dict.setdefault(cnxn_pos_node, [])
                graph_dict[cnxn_pos_node] += [me, shared_vert_nodes[0]]

                if len(shared_vert_nodes) > 1:
                    # connect the second shared vert to the in-between cnxn node
                    graph_dict.setdefault(shared_vert_nodes[1], []).append(cnxn_pos_node)

                    # connect the in-between cnxn_node back to the second shared vert
                    graph_dict[cnxn_pos_node] += [shared_vert_nodes[1]]

        return grph.Graph(
            graph_dict=graph_dict
        )

    def _build_region_graph(
            self,
            regions: Dict[int, Region]
    ) -> grph.Graph:
        logger.debug(f"Building Region Graph...")
        graph_dict = {}

        for id, region in regions.items():
            me = grph.Node(region.node_id, region.pt)

            for neigh, attrs in region.neighbors.items():
                if attrs.Boundary is not None:
                    boundary_pos = attrs.Boundary.Midpoint.as_tuple()
                else:
                    boundary_pos = tuple(float(x) for x in list(attrs.shared_verts.values())[0])

                boundary_node = grph.Node(
                    name=region.neighbors[neigh].cnxn_node_id,
                    pos=boundary_pos
                )

                graph_dict.setdefault(me, []).append(boundary_node)
                graph_dict.setdefault(boundary_node, []).append(me)

        return grph.Graph(
            graph_dict=graph_dict
        )

    def redefine(self,
                 world_bounds: Tuple[int, int],
                 n_regions: int,
                 seed=None):
        if seed is None:
            seed = rnd.randrange(sys.maxsize)
        self._seed = seed
        rnd.seed(self._seed)
        regions, vor, vn_map = _generate_regions(n_regions,
                                                 bounds=[(0, world_bounds[0]),
                                                         (0, world_bounds[1])])
        self.regions: Dict[int, Region] = {x.name: x for x in regions}

        self.village_id = self.choose_village_region().name
        # self.world_graph = self._build_region_graph(self.regions)
        self.world_graph = self._build_boundary_graph(regions=self.regions)

        def __get_villager_or_region_by_id(id: uuid.UUID):
            if id in self.VillagerIdNameMap.keys():
                return self.Villagers[self.VillagerIdNameMap[id]].Position
            if id in self.RegionIdNameMap.keys():
                return self.regions[self.RegionIdNameMap[id]].boundary.exterior.coords

            raise Exception(f'Id: {id} not found in villagers or regions')

        logger.debug(f"Building World SectorTree...")
        self.world_sector_map = SectorTree(
            area_rect=(0, 0, world_bounds[0], world_bounds[1]),
            obj_collider_provider=__get_villager_or_region_by_id,
            shape=(3, 3),
            capacity=1,
            max_lvls=self._sector_tree_depth,
            id='WORLD').add_update_clients(
            {x.id: ClientType.VILLAGER for x in self.Villagers.values()}
        ).add_update_clients({x.id: ClientType.REGION for x in self.regions.values()})

        self._define_rivers()

        self.init_villagers(self._n_init_villagers)

    def _define_rivers(self):
        logger.info(f"Defining Rivers...")
        water_boundaries_node_ids = self._boundaries_between_land_and_water()
        snowpeak_boundaries_node_ids = self._boundaries_between_mountain_and_snowpeak()

        # eval_graph = self.region_graph
        eval_graph = self.world_graph

        if len(water_boundaries_node_ids) == 0 or len(snowpeak_boundaries_node_ids) == 0:
            return

        mouths = [eval_graph.node_by_name(rnd.choice(water_boundaries_node_ids))]
        heads = [eval_graph.node_by_name(rnd.choice(snowpeak_boundaries_node_ids)) for x in range(3)]
        grnd_lvls = {}
        for head in heads:
            grnd_lvls[head] = self.closest_region(
                pt = head.pos,
                excluded_terrain_types=[
                    TerrainType.MOUNTAINS,
                    TerrainType.SNOWPEAK,
                    TerrainType.DEEPSEAWATER,
                    TerrainType.SHALLOWSEAWATER
                ]
            )

        edges_to_add = []

        grnd_cnxn_nodes = []
        for head, clst_grnd in grnd_lvls.items():
            path = eval_graph.astar(
                start=head,
                end=eval_graph.node_by_name(clst_grnd.node_id)
            )
            for e in path.edges[:-1]:
                edges_to_add.append(e)
            grnd_cnxn_nodes.append(edges_to_add[-1].end)

        edges_to_add = set(edges_to_add)


        # Exclude edges that have both neighbor regions ineligible for rivers
        ineligible_cnxns = []
        for id, region in self.regions.items():
            for n_id, attrs in region.neighbors.items():
                if not region.RiverEligible and not self.regions[n_id].RiverEligible:
                    ineligible_cnxns.append(attrs.cnxn_node_id)

        # Also exclude the nodes that define the region's center
        disabled = ineligible_cnxns + [x.name for x in self.world_graph.Nodes.values() if x.name[0] == 'R']

        river_graph = eval_graph.branching_target_seek(
            sources=mouths,
            targets=grnd_cnxn_nodes,
            disabled_node_ids=frozenset(disabled)
        ).reverse()


        river_graph.add_edges(
            edges=list(edges_to_add)
        )

        self.rivers = river_graph


    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _handle_villager_reached_waypoint(self,
                                          v: Villager,
                                          w: wf.Waypoint):
        results = None
        if (w is not None and w.is_destination is True):
            results = self._finish_villager_behavior(v, waypoint=w)
            self._set_villager_behavior(v)
            success = self._handle_villager_behavior(v)

            # if not success:
            #     """CRASH IN"""
            #     self.crash_toggle.toggle()
            #     self.crash_regions = (current_region, region)
            #     """CRASH OUT"""

        return results

    def _set_villager_behavior(self, v: Villager):
        current_behaviors = self.CurrentBehaviors

        if v.Behavior is not None and len(current_behaviors[v.Behavior]) > tp.resolve(self.input_controller.villager_counts[v.Behavior]):
            v.set_behavior(None)

        if v.Behavior is None:
            elgbls = [v for v in BehaviorType if len(current_behaviors[v]) < tp.resolve(self.input_controller.villager_counts[v])]
            if len(elgbls) != 0:
                v.set_behavior(rnd.choice(elgbls))

    def _handle_villager_behavior(self, v: Villager):
        # no behavior, send back to village
        if v.Behavior is None:
            cmn.send_villager_to_region(
                v,
                self.Village,
                villager_current_region_provider=self.get_villager_current_region,
                route_to_waypoints_provider=self.route_to_waypoints,
                route_provider=self.route_between_regions,
            )

            return

        behavior_map = {
            BehaviorType.SCOUT: lambda: scout(
                villager=v,
                villager_current_region_provider=self.get_villager_current_region,
                route_to_waypoints_provider=self.route_to_waypoints,
                route_provider=self.route_between_regions,
                village_region=self.Village,
                regions=self.regions
            ),
            BehaviorType.FARM: lambda: farm(
                villager=v,
                villager_current_region_provider=self.get_villager_current_region,
                route_to_waypoints_provider=self.route_to_waypoints,
                route_provider=self.route_between_regions,
                village_region=self.Village,
                regions=self.regions
            )
        }

        return behavior_map[v.Behavior]()


    def reveal_region_at_waypoint(self, villager: Villager, waypoint: wf.Waypoint) -> BehaviorResults:
        revealed_regions = []

        # Find which regions are connected to the waypoint that was just reached
        cnx_nodes = self.world_graph.adjacent_nodes(
            node=self.world_graph.node_by_name(waypoint.id)
        )
        cnct_regions = {id: r for id, r in self.regions.items() if r.node_id in [x.name for x in cnx_nodes]}

        revealed = next(self.regions[x] for x in cnct_regions
                    if self.regions[x].visibility_level == VisibilityLevel.REVEALING)

        reveal_a_region(r=revealed, regions=self.regions)

        revealed_regions.append(revealed)

        return BehaviorResults(
            villager=villager,
            revealed_regions=revealed_regions
        )


    def _finish_villager_behavior(self, v: Villager, waypoint: wf.Waypoint) -> BehaviorResults:
        if v.Behavior is None:
            return

        behavior_map = {
            BehaviorType.SCOUT: lambda: self.reveal_region_at_waypoint(v, waypoint),
            BehaviorType.FARM: lambda: None
        }

        return behavior_map[v.Behavior]()

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _update_villagers(self, delta_ms: int):
        # Check and add villager
        self.villager_spawn_timer.update(delta_ms=delta_ms)
        if self.villager_spawn_timer.Finished and len(self._villagers) < self._max_villagers:
            self.create_villager()
            self.villager_spawn_timer.reset(start=True)

        ret = []

        # update villagers pos and add new routes if necessary
        for id, v in self._villagers.items():
            reached = v.waypoint_follower.update(delta_time_ms=delta_ms,
                                                 close_enough_tolerance=1,
                                                 slow_enough_tolerance=.1,
                                                 )
            if reached is not None:
                results = self._handle_villager_reached_waypoint(v, reached)
                if results is not None:
                    ret.append(results)

            # if v.waypoint_follower.CurrentTarget is None:
            if v.Behavior is None:
                self._set_villager_behavior(v)
                success = self._handle_villager_behavior(v)

        self.world_sector_map.add_update_clients({x.id: ClientType.VILLAGER for x in self.Villagers.values()})
        return ret

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def __capsule_nearby_clients(self, nearby_range: float, pt: vec.FloatVec):
        return self.world_sector_map.nearby_clients(
            nearby_range,
            pt=pt,
            client_types=[ClientType.REGION]
        )

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def _get_clst_rgn_lgbls(self,
                            pt: vec.FloatVec = None,
                            client_id: UniqueIdentifier = None,
                            nearby_range: int = 50
                            ):
        if pt is None and client_id is None:
            raise Exception(f"at least one of pt and client_id must by non-None")

        # find collocations
        if client_id is not None:
            nearby_region_ids = [x for x in self.world_sector_map.CoLocatedClients[client_id]
                                 if self.world_sector_map.ClientTypeMap[x] == ClientType.REGION]
        else:
            nearby_region_ids = self.__capsule_nearby_clients(
                nearby_range,
                pt=pt,
            )

        return nearby_region_ids

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def closest_region(self,
                       pt: vec.FloatVec = None,
                       client_id: UniqueIdentifier = None,
                       nearby_range: int = 50,
                       excluded_terrain_types: Iterable[TerrainType] = None) -> Region:

        if client_id is not None:
            pt = self.world_sector_map.ClientsPos[client_id]

        nearby_region_ids = self._get_clst_rgn_lgbls(
            pt=pt,
            client_id=client_id,
            nearby_range=nearby_range
        )

        # do formal comparisons and return
        region_candidates: List[Region] = [self.regions[self.RegionIdNameMap[x]] for x in nearby_region_ids
                                           if x in self.RegionIdNameMap.keys()]

        # exclude regions that dont match terrain type
        if excluded_terrain_types is not None:
            region_candidates = [x for x in region_candidates if x.terrain not in excluded_terrain_types]

        min_l = None
        nrst = None
        for r in region_candidates:
            dist = vec.distance_between(pt, r.pt)
            if min_l is None or dist < min_l:
                min_l = dist
                nrst = r

        return nrst

    def update(self, delta_ms: int) -> List[BehaviorResults]:
        if self.crash_toggle.value is False:
            results = self._update_villagers(delta_ms=delta_ms)
            self.world_clock.update(delta_ms)
            return results

    def shuffled_region_generator(self) -> Iterator[Region]:
        regs = list(self.regions.values())
        rnd.shuffle(regs)
        for x in regs:
            yield x

    def _get_villager_name(self) -> int:
        self._last_villager_name += 1
        return self._last_villager_name

    def route_to_waypoints(self,
                           path: Iterable[grph.Node]) -> List[wf.Waypoint]:
        wpts = []
        for ii, seg in enumerate(path):
            if ii == len(path) - 1 and seg.name[0] == 'C' and len(path) > 1:
                pos = vec.interpolate(path[ii-1].pos, path[ii].pos, 0.95)
            else:
                pos = seg.pos

            wpts.append(
                wf.Waypoint(
                    end_pos=pos,
                    id=seg.name,
                    is_destination=True if ii == len(path) - 1 else False
                )
            )

        return wpts

    @timer(logger=logger, log_level=0, time_tracking_class=tt)
    def route_between_regions(self,
                              start: Region,
                              end: Region,
                              ) -> grph.AStarResults:

        disbld = [x.node_id for x in self.regions.values()
                  if x.node_id not in [y.node_id for y in self.NavigateableRegions.values()]]

        disbld += [x.name for x in self.world_graph.Nodes.values() if x.name[0] == 'V']

        if str(end.node_id) in disbld:
            disbld.remove(end.node_id)

        if str(start.node_id) in disbld:
            disbld.remove(start.node_id)

        route = self.world_graph.astar(
            start=self.world_graph.node_by_name(start.node_id),
            end=self.world_graph.node_by_name(end.node_id),
            disabled_node_ids=frozenset(disbld)
        )
        return route

    def get_villager_current_region(self, villager: Villager) -> Region:
        return self.closest_region(client_id=villager.id)

    def _boundaries_between_terrain_types(self,
                                          target_group: Iterable[TerrainType]) -> List[str]:
        ret = []
        for id, region in self.regions.items():
            for n_id in region.neighbors.keys():
                if (self.regions[n_id].terrain in target_group and
                        region.terrain in target_group):
                    # Both are target_group
                    continue

                if (self.regions[n_id].terrain not in target_group and
                        region.terrain not in target_group):
                    # neither are in target_group
                    continue

                if (cnxn := region.neighbors[n_id].cnxn_node_id) not in ret:
                    ret.append(cnxn)

        return ret

    def _boundaries_between_mountain_and_snowpeak(self) -> List[str]:
        return self._boundaries_between_terrain_types([TerrainType.SNOWPEAK])

    def _boundaries_between_land_and_water(self) -> List[str]:
        return self._boundaries_between_terrain_types(
            [TerrainType.DEEPSEAWATER,
              TerrainType.SHALLOWSEAWATER])

    def world_time(self):
        s = self.world_clock.AccumulatedS
        hrs = s / self._seconds_per_hr
        days = int(hrs / 24)
        hrs = int(hrs - 24 * days)

        return days, hrs



    @property
    def Village(self) -> Region:
        return self.regions[self.village_id]

    @property
    def Villagers(self) -> Dict[int, Villager]:
        return self._villagers

    @property
    def NVillagers(self) -> int:
        return len(self.Villagers)

    @property
    def VillagerIdNameMap(self) -> Dict[uuid.UUID, int]:
        return {x.id: x.name for x in self.Villagers.values()}

    @property
    def VillagerNameIdMap(self) -> Dict[int, uuid.UUID]:
        return {x.name: x.id for x in self.Villagers.values()}

    @property
    def RegionIdNameMap(self) -> Dict[uuid.UUID, int]:
        return {x.id: x.name for x in self.regions.values()}

    @property
    def RegionNameIdMap(self) -> Dict[int, uuid.UUID]:
        return {x.name: x.id for x in self.regions.values()}

    @property
    def NavigateableRegions(self) -> Dict[int, Region]:
        return {
            k: v for k, v in self.regions.items() if v.visibility_level in [
                VisibilityLevel.REVEALED
            ] and v.terrain not in [
                 TerrainType.MOUNTAINS,
                 TerrainType.SNOWPEAK,
                 TerrainType.SHALLOWSEAWATER,
                 TerrainType.DEEPSEAWATER
             ]
        }

    @property
    def ReachableRegions(self) -> Dict[int, Region]:
        return {
            k: v for k, v in self.regions.items() if v.visibility_level in [
                VisibilityLevel.REVEALED,
                VisibilityLevel.REVEALABLE,
                VisibilityLevel.REVEALING
            ]
        }

    @property
    def CurrentBehaviors(self) -> Dict[BehaviorType, List[Villager]]:
        ret = {b: [] for b in BehaviorType}

        for id, v in self.Villagers.items():
            ret.setdefault(v.Behavior, []).append(v)
        return ret
